import pytest

from blueetl.config import analysis_model as test_module
from blueetl.utils import load_yaml
from tests.functional.utils import TEST_DATA_PATH as TEST_DATA_PATH_FUNCTIONAL
from tests.unit.utils import TEST_DATA_PATH as TEST_DATA_PATH_UNIT


@pytest.mark.parametrize(
    "config_file",
    [
        pytest.param(f, id=f"{f.relative_to(f.parents[2])}")
        for base in (
            TEST_DATA_PATH_UNIT / "analysis",
            TEST_DATA_PATH_FUNCTIONAL / "sonata" / "config",
            TEST_DATA_PATH_FUNCTIONAL / "bbp" / "config",
        )
        for f in base.glob("*.yaml")
    ],
)
def test_analysis_model(config_file):
    config_dict = load_yaml(config_file)
    config = test_module.MultiAnalysisConfig.model_validate(config_dict)
    assert isinstance(config, test_module.MultiAnalysisConfig)



import pytest
from pandas.testing import assert_frame_equal

from blueetl.store import feather as test_module


@pytest.mark.parametrize(
    "df",
    [
        "storable_df_with_unnamed_index",
        "storable_df_with_named_index",
        "storable_df_with_unnamed_multiindex",
        "storable_df_with_named_multiindex",
    ],
)
def test_dump_load_roundtrip(tmp_path, df, lazy_fixture):
    df = lazy_fixture(df)
    name = "myname"
    store = test_module.FeatherStore(tmp_path)

    store.dump(df, name)
    result = store.load(name)

    assert_frame_equal(result, df)



import pickle
from pathlib import Path

import pytest

from blueetl.adapters import simulation as test_module
from blueetl.adapters.base import AdapterError
from tests.unit.utils import BLUEPY_AVAILABLE, TEST_DATA_PATH, assert_isinstance


@pytest.mark.parametrize(
    "path, population, reports, expected_classes",
    [
        (
            pytest.param(
                "sonata/simulation_config.json",
                "default",
                ["soma_report", "section_report"],
                {
                    "simulation": "bluepysnap.Simulation",
                    "population": "bluepysnap.nodes.NodePopulation",
                    "spikes": "bluepysnap.spike_report.PopulationSpikeReport",
                    "soma_report": "bluepysnap.frame_report.PopulationSomaReport",
                    "section_report": "bluepysnap.frame_report.PopulationCompartmentReport",
                },
                id="snap",
            )
        ),
        (
            pytest.param(
                "bbp/BlueConfig",
                None,
                ["soma", "AllCompartments"],
                {
                    "simulation": "bluepy.Simulation",
                    "population": "bluepy.cells.CellCollection",
                    "spikes": "blueetl.adapters.bluepy.simulation.PopulationSpikesReportImpl",
                    "soma": "blueetl.adapters.bluepy.simulation.PopulationReportImpl",
                    "AllCompartments": "blueetl.adapters.bluepy.simulation.PopulationReportImpl",
                },
                id="bluepy",
                marks=pytest.mark.skipif(not BLUEPY_AVAILABLE, reason="bluepy not available"),
            )
        ),
    ],
)
def test_simulation_adapter(path, population, reports, expected_classes, monkeypatch):
    path = TEST_DATA_PATH / "simulation" / path
    # enter the circuit dir to resolve relative paths in bluepy
    monkeypatch.chdir(path.parent)
    obj = test_module.SimulationAdapter.from_file(path)
    assert_isinstance(obj.instance, expected_classes["simulation"])

    assert obj.exists() is True
    assert obj.is_complete() is True

    # access methods and properties
    pop = obj.circuit.nodes[population]
    assert_isinstance(pop, expected_classes["population"])

    spikes = obj.spikes[population]
    assert_isinstance(spikes, expected_classes["spikes"])

    for report_name in reports:
        report = obj.reports[report_name][population]
        assert_isinstance(report, expected_classes[report_name])

    # test pickle roundtrip
    dumped = pickle.dumps(obj)
    loaded = pickle.loads(dumped)

    assert isinstance(loaded, test_module.SimulationAdapter)
    assert_isinstance(loaded.instance, expected_classes["simulation"])
    # no cached_properties should be loaded after unpickling
    assert sorted(loaded.__dict__) == ["_impl"]
    assert sorted(loaded._impl.__dict__) == ["_simulation"]


def test_simulation_adapter_with_nonexistent_path():
    path = Path("path/to/simulation_config.json")
    obj = test_module.SimulationAdapter.from_file(path)

    assert obj.instance is None
    assert obj.exists() is False
    assert obj.is_complete() is False

    with pytest.raises(AdapterError, match="The implementation doesn't exist"):
        _ = obj.circuit

    with pytest.raises(AdapterError, match="The implementation doesn't exist"):
        _ = obj.spikes

    with pytest.raises(AdapterError, match="The implementation doesn't exist"):
        _ = obj.reports

    # test pickle roundtrip
    dumped = pickle.dumps(obj)
    loaded = pickle.loads(dumped)

    assert isinstance(loaded, test_module.SimulationAdapter)
    assert loaded.instance is None



import pytest

from blueetl.config import analysis as test_module
from blueetl.config.analysis_model import FeaturesConfig, MultiAnalysisConfig
from blueetl.utils import load_yaml
from tests.functional.utils import TEST_DATA_PATH as TEST_DATA_PATH_FUNCTIONAL
from tests.unit.utils import TEST_DATA_PATH as TEST_DATA_PATH_UNIT
from tests.unit.utils import assert_not_duplicates


@pytest.mark.parametrize(
    "config_list, expected_list",
    [
        (
            [],
            [],
        ),
        (
            [
                {
                    "type": "multi",
                    "groupby": ["simulation_id", "circuit_id", "neuron_class", "window"],
                    "function": "module.func",
                    "params": {"PSD": {}},
                    "params_product": {
                        "ratio": [0.25, 0.5],
                        "CPDF": [
                            {"params": {"bin_size": 1}},
                            {"params": {"bin_size": 2}},
                        ],
                    },
                    "params_zip": {
                        "param1": [10, 20],
                        "param2": [11, 21],
                    },
                }
            ],
            [
                {
                    "function": "module.func",
                    "groupby": ["simulation_id", "circuit_id", "neuron_class", "window"],
                    "params": {
                        "CPDF": {"params": {"bin_size": 1}},
                        "PSD": {},
                        "param1": 10,
                        "param2": 11,
                        "ratio": 0.25,
                    },
                    "suffix": "_0_0__0",
                    "type": "multi",
                },
                {
                    "function": "module.func",
                    "groupby": ["simulation_id", "circuit_id", "neuron_class", "window"],
                    "params": {
                        "CPDF": {"params": {"bin_size": 1}},
                        "PSD": {},
                        "param1": 20,
                        "param2": 21,
                        "ratio": 0.25,
                    },
                    "suffix": "_0_0__1",
                    "type": "multi",
                },
                {
                    "function": "module.func",
                    "groupby": ["simulation_id", "circuit_id", "neuron_class", "window"],
                    "params": {
                        "CPDF": {"params": {"bin_size": 2}},
                        "PSD": {},
                        "param1": 10,
                        "param2": 11,
                        "ratio": 0.25,
                    },
                    "suffix": "_0_1__0",
                    "type": "multi",
                },
                {
                    "function": "module.func",
                    "groupby": ["simulation_id", "circuit_id", "neuron_class", "window"],
                    "params": {
                        "CPDF": {"params": {"bin_size": 2}},
                        "PSD": {},
                        "param1": 20,
                        "param2": 21,
                        "ratio": 0.25,
                    },
                    "suffix": "_0_1__1",
                    "type": "multi",
                },
                {
                    "function": "module.func",
                    "groupby": ["simulation_id", "circuit_id", "neuron_class", "window"],
                    "params": {
                        "CPDF": {"params": {"bin_size": 1}},
                        "PSD": {},
                        "param1": 10,
                        "param2": 11,
                        "ratio": 0.5,
                    },
                    "suffix": "_1_0__0",
                    "type": "multi",
                },
                {
                    "function": "module.func",
                    "groupby": ["simulation_id", "circuit_id", "neuron_class", "window"],
                    "params": {
                        "CPDF": {"params": {"bin_size": 1}},
                        "PSD": {},
                        "param1": 20,
                        "param2": 21,
                        "ratio": 0.5,
                    },
                    "suffix": "_1_0__1",
                    "type": "multi",
                },
                {
                    "function": "module.func",
                    "groupby": ["simulation_id", "circuit_id", "neuron_class", "window"],
                    "params": {
                        "CPDF": {"params": {"bin_size": 2}},
                        "PSD": {},
                        "param1": 10,
                        "param2": 11,
                        "ratio": 0.5,
                    },
                    "suffix": "_1_1__0",
                    "type": "multi",
                },
                {
                    "function": "module.func",
                    "groupby": ["simulation_id", "circuit_id", "neuron_class", "window"],
                    "params": {
                        "CPDF": {"params": {"bin_size": 2}},
                        "PSD": {},
                        "param1": 20,
                        "param2": 21,
                        "ratio": 0.5,
                    },
                    "suffix": "_1_1__1",
                    "type": "multi",
                },
            ],
        ),
    ],
)
def test_config__resolve_features(config_list, expected_list):
    config_list = [FeaturesConfig(**d) for d in config_list]
    expected_list = [FeaturesConfig(**d) for d in expected_list]

    result = test_module._resolve_features(config_list)

    assert result == expected_list
    assert_not_duplicates(result)


def test_config__resolve_features_error():
    config_list = [
        {
            "type": "multi",
            "groupby": ["simulation_id", "circuit_id", "neuron_class", "window"],
            "function": "module.func",
            "params_zip": {
                "param1": [10, 20],
                "param2": [11, 21, 31],
            },
        }
    ]
    config_list = [FeaturesConfig(**d) for d in config_list]

    with pytest.raises(ValueError, match="All the zip params must have the same length"):
        test_module._resolve_features(config_list)


@pytest.mark.parametrize(
    "config_file",
    [
        pytest.param(f, id=f"{f.relative_to(f.parents[2])}")
        for base in (
            TEST_DATA_PATH_UNIT / "analysis",
            TEST_DATA_PATH_FUNCTIONAL / "sonata" / "config",
            TEST_DATA_PATH_FUNCTIONAL / "bbp" / "config",
        )
        for f in base.glob("*.yaml")
    ],
)
def test_init_multi_analysis_configuration(config_file):
    config_dict = load_yaml(config_file)
    base_path = config_file.parent
    result = test_module.init_multi_analysis_configuration(
        config_dict, base_path=base_path, extra_params={}
    )
    assert isinstance(result, MultiAnalysisConfig)



"""Repository."""

import logging
import warnings
from abc import ABC, abstractmethod
from functools import cached_property
from typing import Any, Generic, Optional

import pandas as pd

from blueetl.cache import CacheManager
from blueetl.campaign.config import SimulationCampaign
from blueetl.config.analysis_model import ExtractionConfig
from blueetl.constants import CIRCUIT_ID, SIMULATION_ID, SIMULATION_PATH
from blueetl.extract.base import ExtractorT
from blueetl.extract.compartment_report import CompartmentReport
from blueetl.extract.neuron_classes import NeuronClasses
from blueetl.extract.neurons import Neurons
from blueetl.extract.report import ReportExtractor
from blueetl.extract.simulations import Simulations
from blueetl.extract.soma_report import SomaReport
from blueetl.extract.spikes import Spikes
from blueetl.extract.windows import Windows
from blueetl.resolver import Resolver
from blueetl.utils import timed

L = logging.getLogger(__name__)


class BaseExtractor(ABC, Generic[ExtractorT]):
    """BaseExtractor class."""

    def __init__(self, repo: "Repository") -> None:
        """Initialize the object."""
        self._repo = repo

    @abstractmethod
    def extract_new(self) -> ExtractorT:
        """Instantiate an object from the configuration."""

    @abstractmethod
    def extract_cached(self, df: pd.DataFrame, name: str) -> ExtractorT:
        """Instantiate an object from a cached DataFrame."""

    def extract(self, name: str) -> ExtractorT:
        """Return an object extracted from the cache or as new.

        Args:
            name: name of the dataframe.
        """
        with timed(L.debug, f"Extracting {name}") as messages:
            is_cached = self._repo.cache_manager.is_repo_cached(name)
            if is_cached:
                df = self._repo.cache_manager.load_repo(name)
                instance = self.extract_cached(df, name)
            else:
                instance = self.extract_new()
            assert instance is not None, "The extraction didn't return a valid instance."
            assert is_cached is instance._cached  # pylint: disable=protected-access
            is_filtered = instance._filtered  # pylint: disable=protected-access
            if not is_cached or is_filtered:
                with timed(L.info, f"Writing cached {name}"):
                    self._repo.cache_manager.dump_repo(df=instance.to_pandas(), name=name)
            messages[:] = [f"Extracted {name}: {is_cached=} {is_filtered=} rows={len(instance.df)}"]
            return instance


class SimulationsExtractor(BaseExtractor[Simulations]):
    """SimulationsExtractor class."""

    def extract_new(self) -> Simulations:
        """Instantiate an object from the configuration."""
        return Simulations.from_config(
            config=self._repo.simulations_config,
            query=self._repo.simulations_filter,
        )

    def extract_cached(self, df: pd.DataFrame, name: str) -> Simulations:
        """Instantiate an object from a cached DataFrame."""
        query = None
        if self._repo.needs_filter(name):
            query = self._repo.simulations_filter
        return Simulations.from_pandas(df, query=query, cached=True)


class NeuronsExtractor(BaseExtractor[Neurons]):
    """NeuronsExtractor class."""

    def extract_new(self) -> Neurons:
        """Instantiate an object from the configuration."""
        return Neurons.from_simulations(
            simulations=self._repo.simulations,
            neuron_classes=self._repo.extraction_config.neuron_classes,
        )

    def extract_cached(self, df: pd.DataFrame, name: str) -> Neurons:
        """Instantiate an object from a cached DataFrame."""
        query = None
        if self._repo.needs_filter(name):
            selected_sims = self._repo.simulations.df.etl.q(simulation_id=self._repo.simulation_ids)
            query = {CIRCUIT_ID: sorted(set(selected_sims[CIRCUIT_ID]))}
        return Neurons.from_pandas(df, query=query, cached=True)


class NeuronClassesExtractor(BaseExtractor[NeuronClasses]):
    """NeuronClassesExtractor class."""

    def extract_new(self) -> NeuronClasses:
        """Instantiate an object from the configuration."""
        return NeuronClasses.from_neurons(
            neurons=self._repo.neurons, neuron_classes=self._repo.extraction_config.neuron_classes
        )

    def extract_cached(self, df: pd.DataFrame, name: str) -> NeuronClasses:
        """Instantiate an object from a cached DataFrame."""
        query = None
        if self._repo.needs_filter(name):
            selected_sims = self._repo.simulations.df.etl.q(simulation_id=self._repo.simulation_ids)
            query = {CIRCUIT_ID: sorted(set(selected_sims[CIRCUIT_ID]))}
        return NeuronClasses.from_pandas(df, query=query, cached=True)


class WindowsExtractor(BaseExtractor[Windows]):
    """WindowsExtractor class."""

    def extract_new(self) -> Windows:
        """Instantiate an object from the configuration."""
        assert self._repo.resolver is not None
        return Windows.from_simulations(
            simulations=self._repo.simulations,
            windows_config=self._repo.extraction_config.windows,
            trial_steps_config=self._repo.extraction_config.trial_steps,
            resolver=self._repo.resolver,
        )

    def extract_cached(self, df: pd.DataFrame, name: str) -> Windows:
        """Instantiate an object from a cached DataFrame."""
        query = None
        if self._repo.needs_filter(name):
            query = {SIMULATION_ID: self._repo.simulation_ids}
        return Windows.from_pandas(df, query=query, cached=True)


class SpikesExtractor(BaseExtractor[Spikes]):
    """SpikesExtractor class."""

    def extract_new(self) -> Spikes:
        """Instantiate an object from the configuration."""
        return Spikes.from_simulations(
            simulations=self._repo.simulations,
            neurons=self._repo.neurons,
            windows=self._repo.windows,
            neuron_classes=self._repo.neuron_classes,
            name=self._repo.extraction_config.report.name,
        )

    def extract_cached(self, df: pd.DataFrame, name: str) -> Spikes:
        """Instantiate an object from a cached DataFrame."""
        query = None
        if self._repo.needs_filter(name):
            query = {SIMULATION_ID: self._repo.simulation_ids}
        return Spikes.from_pandas(df, query=query, cached=True)


class SomaReportExtractor(BaseExtractor[SomaReport]):
    """SomaReportExtractor class."""

    def extract_new(self) -> SomaReport:
        """Instantiate an object from the configuration."""
        return SomaReport.from_simulations(
            simulations=self._repo.simulations,
            neurons=self._repo.neurons,
            windows=self._repo.windows,
            neuron_classes=self._repo.neuron_classes,
            name=self._repo.extraction_config.report.name,
        )

    def extract_cached(self, df: pd.DataFrame, name: str) -> SomaReport:
        """Instantiate an object from a cached DataFrame."""
        query = None
        if self._repo.needs_filter(name):
            query = {SIMULATION_ID: self._repo.simulation_ids}
        return SomaReport.from_pandas(df, query=query, cached=True)


class CompartmentReportExtractor(BaseExtractor[CompartmentReport]):
    """CompartmentReportExtractor class."""

    def extract_new(self) -> CompartmentReport:
        """Instantiate an object from the configuration."""
        return CompartmentReport.from_simulations(
            simulations=self._repo.simulations,
            neurons=self._repo.neurons,
            windows=self._repo.windows,
            neuron_classes=self._repo.neuron_classes,
            name=self._repo.extraction_config.report.name,
        )

    def extract_cached(self, df: pd.DataFrame, name: str) -> CompartmentReport:
        """Instantiate an object from a cached DataFrame."""
        query = None
        if self._repo.needs_filter(name):
            query = {SIMULATION_ID: self._repo.simulation_ids}
        return CompartmentReport.from_pandas(df, query=query, cached=True)


class Repository:
    """Repository class."""

    def __init__(
        self,
        simulations_config: SimulationCampaign,
        extraction_config: ExtractionConfig,
        cache_manager: CacheManager,
        simulations_filter: Optional[dict[str, Any]] = None,
        resolver: Optional[Resolver] = None,
    ) -> None:
        """Initialize the repository.

        Args:
            simulations_config: simulation campaign configuration.
            extraction_config: extraction configuration.
            cache_manager: cache manager responsible to load and dump dataframes.
            simulations_filter: optional simulations filter.
            resolver: resolver instance.
        """
        self._extraction_config = extraction_config
        self._simulations_config = simulations_config
        self._cache_manager = cache_manager
        self._simulations_filter = simulations_filter
        self._resolver = resolver
        report_type = extraction_config.report.type
        available_reports: dict[str, type[BaseExtractor]] = {
            "spikes": SpikesExtractor,
            "soma": SomaReportExtractor,
            "compartment": CompartmentReportExtractor,
        }
        self._mapping: dict[str, type[BaseExtractor]] = {
            "simulations": SimulationsExtractor,
            "neurons": NeuronsExtractor,
            "neuron_classes": NeuronClassesExtractor,
            "windows": WindowsExtractor,
            "report": available_reports[report_type],
        }
        self._names = list(self._mapping)

    def __getstate__(self) -> dict:
        """Get the object state when the object is pickled."""
        if not self.is_cached():
            # ensure that the dataframes are extracted and stored to disk,
            # because we want to be able to use the cached data in the subprocesses.
            L.info("Extracting dataframes before serialization")
            self.extract()
        # Copy the object's state, excluding the unpicklable entries.
        names_set = set(self.names)
        return {k: v for k, v in self.__dict__.items() if k not in names_set}

    def __setstate__(self, state: dict) -> None:
        """Set the object state when the object is unpickled."""
        self.__dict__.update(state)

    @property
    def extraction_config(self) -> ExtractionConfig:
        """Access to the extraction configuration."""
        return self._extraction_config

    @property
    def simulations_config(self) -> SimulationCampaign:
        """Access to the simulation campaign configuration."""
        return self._simulations_config

    @property
    def cache_manager(self) -> CacheManager:
        """Access to the cache manager."""
        return self._cache_manager

    @property
    def simulations_filter(self) -> Optional[dict[str, Any]]:
        """Access to the simulations filter."""
        return self._simulations_filter

    @property
    def names(self) -> list[str]:
        """Return the list of names of the extracted objects."""
        return self._names

    @property
    def resolver(self) -> Optional[Resolver]:
        """Return the resolver."""
        return self._resolver

    @cached_property
    def simulations(self) -> Simulations:
        """Return the Simulations extraction."""
        return self._mapping["simulations"](self).extract(name="simulations")

    @cached_property
    def neurons(self) -> Neurons:
        """Return the Neurons extraction."""
        return self._mapping["neurons"](self).extract(name="neurons")

    @cached_property
    def neuron_classes(self) -> NeuronClasses:
        """Return the NeuronClasses extraction."""
        return self._mapping["neuron_classes"](self).extract(name="neuron_classes")

    @cached_property
    def windows(self) -> Windows:
        """Return the Windows extraction."""
        return self._mapping["windows"](self).extract(name="windows")

    @property
    def spikes(self) -> ReportExtractor:
        """Return the Spikes extraction."""
        warnings.warn(
            "Accessing Repository.spikes is deprecated, please use Repository.report",
            FutureWarning,
            stacklevel=2,
        )
        assert isinstance(self.report, Spikes)
        return self.report

    @cached_property
    def report(self) -> ReportExtractor:
        """Return the Report extraction."""
        return self._mapping["report"](self).extract(name="report")

    @property
    def simulation_ids(self) -> list[int]:
        """Return the list of simulation ids, possibly filtered."""
        return self.simulations.df[SIMULATION_ID].to_list()

    def extract(self) -> None:
        """Extract all the dataframes."""
        for name in self.names:
            getattr(self, name)
        self.check_extractions()

    def is_extracted(self) -> bool:
        """Return True if all the dataframes have been extracted and loaded in memory."""
        # note: the cached_property is stored as an attribute after it's accessed
        return all(name in self.__dict__ for name in self.names)

    def check_extractions(self) -> None:
        """Check that all the dataframes have been extracted."""
        if not self.is_extracted():
            raise RuntimeError("Not all the dataframes have been extracted")

    def is_cached(self) -> bool:
        """Return True if all the dataframes have been cached (loaded in memory or not)."""
        return all(self.cache_manager.is_repo_cached(name) for name in self.names)

    def missing_simulations(self) -> pd.DataFrame:
        """Return a DataFrame with the simulations ignored because of missing spikes.

        Returns:
            pd.DataFrame with the simulation conditions and simulation_path as columns,
                and one record for each ignored and missing simulation.
        """
        all_simulations = self._simulations_config.get()
        extracted_simulations = self.simulations.df[[SIMULATION_PATH]]
        return (
            pd.merge(
                all_simulations,
                extracted_simulations,
                left_on=[SIMULATION_PATH],
                right_on=[SIMULATION_PATH],
                how="left",
                indicator=True,
            )
            .etl.q(_merge="left_only")
            .drop(columns="_merge")
        )

    def show(self) -> None:
        """Print some information about the instance, mainly for debug and inspection."""
        for name in self.names:
            print("~" * 80)
            print("Extraction:", name)
            print(getattr(self, name).df)

    def apply_filter(self, simulations_filter: dict[str, Any]) -> "Repository":
        """Apply the given filter and return a new object."""
        return FilteredRepository(parent=self, simulations_filter=simulations_filter)

    def needs_filter(self, name: str) -> bool:
        """Return True if the repository needs to be filtered during the extraction."""
        return bool(self.simulations_filter) and self.cache_manager.repo_cache_needs_filter(name)


class FilteredRepository(Repository):
    """FilteredRepository class."""

    def __init__(self, parent: Repository, simulations_filter: dict[str, Any]) -> None:
        """Initialize the object using the given dict of DataFrames.

        Filtered dataframes are never written to disk.
        """
        super().__init__(
            simulations_config=parent.simulations_config,
            extraction_config=parent.extraction_config,
            cache_manager=parent.cache_manager.to_readonly(),
            simulations_filter=simulations_filter,
        )
        dataframes = {name: getattr(parent, name).df for name in parent.names}
        self._assign_from_dataframes(dataframes)

    def _assign_from_dataframes(self, dicts: dict[str, pd.DataFrame]) -> None:
        """Assign the repository properties from the given dict of DataFrames."""
        for name, df in dicts.items():
            assert name not in self.__dict__
            value = self._mapping[name](self).extract_cached(df, name)
            setattr(self, name, value)

    def needs_filter(self, name: str) -> bool:
        """Return True if the repository needs to be filtered during the extraction."""
        return bool(self.simulations_filter)



"""Adapters for Simulation."""

from collections.abc import Mapping
from pathlib import Path
from typing import Optional

from blueetl.adapters.base import BaseAdapter
from blueetl.adapters.circuit import CircuitAdapter
from blueetl.adapters.interfaces.simulation import (
    PopulationReportInterface,
    PopulationSpikesReportInterface,
    SimulationInterface,
)


class SimulationAdapter(BaseAdapter[SimulationInterface]):
    """Simulation Adapter."""

    @classmethod
    def from_file(cls, filepath: Optional[Path]) -> "SimulationAdapter":
        """Load and return a new object from file."""
        # pylint: disable=import-outside-toplevel
        if not filepath or not filepath.exists():
            return cls(None)
        SimulationImpl: type[SimulationInterface]
        if filepath.suffix == ".json":
            from blueetl.adapters.impl.bluepysnap.simulation import Simulation, SimulationImpl
        else:
            from blueetl.adapters.impl.bluepy.simulation import Simulation, SimulationImpl
        impl = SimulationImpl(Simulation(str(filepath)))
        return cls(impl)

    def is_complete(self) -> bool:
        """Return True if the simulation is complete, False otherwise."""
        return self.exists() and self._ensure_impl.is_complete()

    @property
    def circuit(self) -> CircuitAdapter:
        """Return the circuit used for the simulation."""
        return CircuitAdapter(self._ensure_impl.circuit)

    @property
    def spikes(self) -> Mapping[Optional[str], PopulationSpikesReportInterface]:
        """Return the spikes report as a dict: population -> report."""
        return self._ensure_impl.spikes

    @property
    def reports(self) -> Mapping[str, Mapping[Optional[str], PopulationReportInterface]]:
        """Return the reports as a dict: name -> population -> report."""
        return self._ensure_impl.reports



"""Features adapted from BlueNetworkActivityComparison/bnac/data_processor.py."""

import logging

import numpy as np
import pandas as pd
from blueetl_core.utils import smart_concat
from scipy.ndimage import gaussian_filter

from blueetl.constants import BIN, COUNT, GID, NEURON_CLASS_INDEX, TIME, TIMES, TRIAL

L = logging.getLogger(__name__)
FIRST = "first"


def _get_initial_spiking_stats(repo, key, df, params):
    # pylint: disable=unused-argument
    duration = repo.windows.get_duration(key.window)

    # df with index (trial, gid) and columns (count, times)
    spikes_by_trial = df.groupby([TRIAL, GID, NEURON_CLASS_INDEX])[TIME].agg(
        **{
            COUNT: "count",
            FIRST: "min",
            TIMES: lambda x: [i for i in x if not np.isnan(i)],  # slow
        }
    )
    # first spike for each trial and gid, averaged across all trials where the neuron was present
    first_spike_time_means_cort_zeroed = (
        spikes_by_trial[FIRST]
        .groupby([GID, NEURON_CLASS_INDEX])
        .mean()
        .rename("first_spike_time_means_cort_zeroed")
    )

    mean_spike_counts = spikes_by_trial[COUNT].fillna(0).groupby([GID, NEURON_CLASS_INDEX]).mean()
    mean_spike_counts = mean_spike_counts.rename("mean_spike_counts")
    mean_of_spike_counts_for_each_trial = (
        spikes_by_trial[COUNT]
        .fillna(0)
        .groupby(TRIAL)
        .mean()
        .rename("mean_of_spike_counts_for_each_trial")
    )

    mean_firing_rates_per_second = mean_spike_counts * 1000.0 / duration
    mean_firing_rates_per_second = mean_firing_rates_per_second.rename(
        "mean_firing_rates_per_second"
    )

    return {
        # by_gid_and_trial
        "spikes_by_trial": spikes_by_trial,
        # by_gid
        "first_spike_time_means_cort_zeroed": first_spike_time_means_cort_zeroed,
        "mean_spike_counts": mean_spike_counts,
        "mean_firing_rates_per_second": mean_firing_rates_per_second,
        # by_neuron_class_and_trial
        "mean_of_spike_counts_for_each_trial": mean_of_spike_counts_for_each_trial,
        # by_neuron_class (scalar values)
        "mean_of_mean_spike_counts": np.nanmean(mean_spike_counts),
        "mean_of_mean_firing_rates_per_second": np.mean(mean_firing_rates_per_second),
        "std_of_mean_firing_rates_per_second": np.std(mean_firing_rates_per_second),
    }


def _get_histogram_features(repo, key, df, params):
    # pylint: disable=unused-argument,too-many-locals
    number_of_trials = repo.windows.get_number_of_trials(key.window)
    duration = repo.windows.get_duration(key.window)
    t_start, t_stop = repo.windows.get_bounds(key.window)
    # all the spike times are concatenated regardless of the trial
    times = df[TIME].to_numpy()
    hist, _ = np.histogram(times, range=[t_start, t_stop], bins=int(duration))
    num_target_cells = len(
        repo.neurons.df.etl.q(circuit_id=key.circuit_id, neuron_class=key.neuron_class)
    )
    hist = hist / (num_target_cells * number_of_trials)
    min_hist = np.min(hist)
    max_hist = np.max(hist)
    norm_hist = hist / (max_hist or 1)
    smoothed_hist = gaussian_filter(hist, sigma=4.0)
    max_smoothed_hist = np.max(smoothed_hist)
    norm_smoothed_hist = smoothed_hist / (max_smoothed_hist or 1)
    return {
        "spike_times_normalised_hist_1ms_bin": hist,
        "mean_of_spike_times_normalised_hist_1ms_bin": np.mean(hist),
        "min_of_spike_times_normalised_hist_1ms_bin": min_hist,
        "max_of_spike_times_normalised_hist_1ms_bin": max_hist,
        "argmax_spike_times_hist_1ms_bin": np.argmax(hist),
        "spike_times_max_normalised_hist_1ms_bin": norm_hist,
        "smoothed_3ms_spike_times_max_normalised_hist_1ms_bin": norm_smoothed_hist,
    }


def calculate_features_multi(repo, key, df, params):
    """Calculate multiple features at the same time."""
    export_all_neurons = params.get("export_all_neurons", False)
    spiking_stats = _get_initial_spiking_stats(repo, key, df, params)
    histogram_features = _get_histogram_features(repo, key, df, params)

    # df with (gid) as index, and features as columns
    by_gid = smart_concat(
        [
            spiking_stats["first_spike_time_means_cort_zeroed"],
            spiking_stats["mean_spike_counts"],
            spiking_stats["mean_firing_rates_per_second"],
        ],
        axis=1,
    )
    if not export_all_neurons:
        # return only neurons with spikes
        # TODO: to be tested
        by_gid = by_gid.dropna(how="all")

    # df with (trial, gid) as index, and features as columns
    by_gid_and_trial = smart_concat(
        [
            spiking_stats["spikes_by_trial"],
        ],
        axis=1,
    )
    if not export_all_neurons:
        # return only neurons with spikes
        # TODO: to be tested
        by_gid_and_trial = by_gid_and_trial.dropna(how="all")

    # df with features as columns and a single row
    # the index will be dropped when concatenating because it's unnamed
    by_neuron_class = pd.DataFrame(
        {
            "mean_of_mean_spike_counts": spiking_stats["mean_of_mean_spike_counts"],
            "mean_of_mean_firing_rates_per_second": spiking_stats[
                "mean_of_mean_firing_rates_per_second"
            ],
            "std_of_mean_firing_rates_per_second": spiking_stats[
                "std_of_mean_firing_rates_per_second"
            ],
            "mean_of_spike_times_normalised_hist_1ms_bin": histogram_features[
                "mean_of_spike_times_normalised_hist_1ms_bin"
            ],
            "min_of_spike_times_normalised_hist_1ms_bin": histogram_features[
                "min_of_spike_times_normalised_hist_1ms_bin"
            ],
            "max_of_spike_times_normalised_hist_1ms_bin": histogram_features[
                "max_of_spike_times_normalised_hist_1ms_bin"
            ],
            "argmax_spike_times_hist_1ms_bin": histogram_features[
                "argmax_spike_times_hist_1ms_bin"
            ],
        },
        index=[0],
    )

    # df with (trial) as index, and features as columns
    by_neuron_class_and_trial = spiking_stats["mean_of_spike_counts_for_each_trial"].to_frame()

    # df with (bin) as index, and features as columns
    histograms = pd.DataFrame(
        {
            "spike_times_normalised_hist_1ms_bin": histogram_features[
                "spike_times_normalised_hist_1ms_bin"
            ],
            "spike_times_max_normalised_hist_1ms_bin": histogram_features[
                "spike_times_max_normalised_hist_1ms_bin"
            ],
            "smoothed_3ms_spike_times_max_normalised_hist_1ms_bin": histogram_features[
                "smoothed_3ms_spike_times_max_normalised_hist_1ms_bin"
            ],
        }
    ).rename_axis(index=BIN)

    return {
        "by_gid": by_gid,
        "by_gid_and_trial": by_gid_and_trial,
        "by_neuron_class": by_neuron_class,
        "by_neuron_class_and_trial": by_neuron_class_and_trial,
        "histograms": histograms,
    }



"""Main CLI."""

import click

from blueetl import __version__
from blueetl.apps.convert import convert_spikes
from blueetl.apps.migrate import migrate_config
from blueetl.apps.run import run
from blueetl.apps.validate import validate_config


class NaturalOrderGroup(click.Group):
    """Click group preserving the order of commands."""

    def list_commands(self, ctx: click.Context) -> list[str]:
        """Return the list of possible commands."""
        return list(self.commands)


@click.group(cls=NaturalOrderGroup)
@click.version_option(__version__)
def cli():
    """The CLI entry point."""


cli.add_command(run)
cli.add_command(migrate_config)
cli.add_command(validate_config)
cli.add_command(convert_spikes)



"""Features collection."""

import logging
import tempfile
from collections import Counter, defaultdict
from collections.abc import Iterator
from copy import deepcopy
from dataclasses import dataclass
from functools import cached_property, partial
from pathlib import Path
from typing import Any, NamedTuple, Optional, Union

import pandas as pd
from blueetl_core.parallel import isolated
from blueetl_core.utils import smart_concat

from blueetl.cache import CacheManager
from blueetl.config.analysis_model import FeaturesConfig
from blueetl.constants import SIMULATION_ID
from blueetl.extract.feature import Feature
from blueetl.parallel import merge_filter
from blueetl.repository import Repository
from blueetl.store.parquet import ParquetStore
from blueetl.utils import (
    all_equal,
    ensure_dtypes,
    extract_items,
    get_shmdir,
    import_by_string,
    timed,
)

L = logging.getLogger(__name__)


class ConcatenatedFeatures:
    """ConcatenatedFeatures class.

    It can be used to view as a single DataFrame all the features calculated
    for various combinations of parameters.
    """

    def __init__(
        self,
        parent: "FeaturesCollection",
        configs: Optional[dict[str, FeaturesConfig]] = None,
    ) -> None:
        """Initialize the object."""
        self._parent = parent
        self._configs: dict[str, FeaturesConfig] = configs or {}

    def clone(self, parent: "FeaturesCollection") -> "ConcatenatedFeatures":
        """Return a copy using the same config and the new parent."""
        return self.__class__(parent, deepcopy(self._configs))

    def update(self, feature_name: str, features_config: FeaturesConfig) -> None:
        """Update the list of features and configurations."""
        assert feature_name not in self._configs, f"Duplicate feature_name: {feature_name}"
        self._configs[feature_name] = features_config
        self.clear_cache()

    def clear_cache(self) -> None:
        """Clear the cached properties."""
        for key in "params", "aliases", "df":
            self.__dict__.pop(key, None)

    @cached_property
    def params(self) -> pd.DataFrame:
        """Return all the parameters as a DataFrame."""
        return pd.DataFrame(
            [
                dict(extract_items(config.params))
                for params_id, config in enumerate(self._configs.values())
            ],
            index=pd.RangeIndex(len(self._configs), name="params_id"),
        )

    @cached_property
    def aliases(self) -> pd.DataFrame:
        """Return the varying column names used as parameters and their aliases."""
        params_df = self.params
        # drop columns containing only constant values
        columns = [col for col in params_df.columns if not all_equal(params_df[col])]
        # shorten unambiguous column names
        aliases = [col.rsplit(".", maxsplit=1)[-1] for col in columns]
        counter = Counter(aliases)
        return pd.DataFrame(
            [
                {"column": col, "alias": alias if counter[alias] == 1 else col}
                for col, alias in zip(columns, aliases)
            ]
        )

    @cached_property
    def df(self) -> pd.DataFrame:
        """Return the concatenation of features."""
        params_df = self.params[self.aliases["column"]]
        params_df.columns = self.aliases["alias"]
        # add params_id as a column
        params_df = params_df.reset_index()
        # concatenate all the features together
        return smart_concat(
            self._augment_dataframe(self._partial_df(name), params)
            for name, (_, params) in zip(self._configs, params_df.etl.iterdict())
        )

    def _partial_df(self, name: str) -> pd.DataFrame:
        """Return the specified partial DataFrame from the parent."""
        return getattr(self._parent, name).df

    @staticmethod
    def _augment_dataframe(df: pd.DataFrame, params: dict[str, Any]) -> pd.DataFrame:
        """Return a copy of the DataFrame after adding columns from the given dict."""
        # ensure that a value can be assigned to a cell even when it's a list
        params = {k: [v] * len(df) for k, v in params.items()}
        return df.assign(**params)


@dataclass(frozen=True)
class FeaturesConfigKey:
    """Group of common keys used to build a group of FeaturesConfig.

    FeaturesConfig in the same group are grouped and processed together.

    Warning: the keys are considered immutable, and must not be modified after the object creation.
    """

    groupby: list[str]
    neuron_classes: list[str]
    windows: list[str]

    @classmethod
    def from_config(cls, config: FeaturesConfig) -> "FeaturesConfigKey":
        """Return an instance from the given FeaturesConfig."""
        return cls(
            groupby=config.groupby,
            neuron_classes=config.neuron_classes,
            windows=config.windows,
        )

    def __hash__(self):
        """Return the hash of the object."""
        return hash((tuple(self.groupby), tuple(self.neuron_classes), tuple(self.windows)))


class FeaturesCollection:
    """FeaturesCollection class."""

    def __init__(
        self,
        features_configs: list[FeaturesConfig],
        repo: Repository,
        cache_manager: CacheManager,
    ) -> None:
        """Initialize the FeaturesCollection from the given list of configurations.

        Args:
            features_configs: list of features configuration dicts.
            repo: Repository instance.
            cache_manager: CacheManager instance.
        """
        self._features_configs = features_configs
        self._repo = repo
        self._cache_manager = cache_manager
        self._data: dict[str, Feature] = {}
        self._concatenated_features: dict[str, ConcatenatedFeatures] = {}

    @property
    def names(self) -> list[str]:
        """Return the names of all the calculated features."""
        if not self._data:
            self.calculate()
        return sorted([*self._data, *self._concatenated_features])

    @property
    def cache_manager(self) -> CacheManager:
        """Access to the cache manager."""
        return self._cache_manager

    def __getattr__(self, name: str) -> Union[Feature, ConcatenatedFeatures]:
        """Return the features by name.

        It mimics the behaviour of Repository, where each Extractor can be accessed by attribute.

        Args:
            name: name of the features.
        """
        if name in self._data:
            return self._data[name]
        if name in self._concatenated_features:
            return self._concatenated_features[name]
        raise AttributeError(f"{self.__class__.__name__!r} object has no attribute {name!r}")

    def __dir__(self):
        """Allow autocompletion of dynamic attributes."""
        return list(super().__dir__()) + list(self._data) + list(self._concatenated_features)

    def _update_features(self, mapping: dict[str, Feature]) -> None:
        if overlapping := set(mapping).intersection(self._data):
            raise RuntimeError(
                f"Overlapping features dataframes aren't allowed: {sorted(overlapping)}"
            )
        if overlapping := set(mapping).intersection(self._concatenated_features):
            raise RuntimeError(
                f"Features cannot overlap with concatenated features: {sorted(overlapping)}"
            )
        self._data.update(mapping)

    def _update_concatenated_features(
        self, feature_names: list[str], features_config: FeaturesConfig
    ) -> None:
        if not features_config.suffix:
            # not a concatenated feature, nothing to do
            return
        for name in feature_names:
            basename = name.removesuffix(features_config.suffix)
            if basename in self._data:
                raise RuntimeError(
                    f"Concatenated features {basename} cannot overlap with features dataframes"
                )
            if basename not in self._concatenated_features:
                self._concatenated_features[basename] = ConcatenatedFeatures(self)
            concatenated = self._concatenated_features[basename]
            concatenated.update(feature_name=name, features_config=features_config)

    def show(self) -> None:
        """Print some information about the instance, mainly for debug and inspection."""
        for name in self.names:
            print("~" * 80)
            print("Features:", name)
            print(getattr(self, name).df)

    def calculate(self) -> None:
        """Calculate all the features based on the configuration."""
        if self._data:
            # Features already calculated
            return

        def _process_features(
            features_config: FeaturesConfig, features: dict[str, Feature]
        ) -> None:
            """Update the features of the instance, and write the cache if needed."""
            self._update_features(features)
            self._update_concatenated_features(list(features), features_config)
            to_be_written: dict[str, pd.DataFrame] = {}
            for name, f in features.items():
                if not f._cached or f._filtered:  # pylint: disable=protected-access
                    to_be_written[name] = f.to_pandas()
            if to_be_written:
                with timed(L.info, "Writing cached features"):
                    self.cache_manager.dump_features(to_be_written, features_config=features_config)

        def _log_features(features: dict[str, Feature], n: int, tot: int, features_id) -> None:
            """Log a message about the features being processed."""
            msg = "\n".join(
                # pylint: disable=protected-access
                f"- {name}: cached={f._cached}, filtered={f._filtered}"
                for name, f in features.items()
            )
            L.info("Calculated features %s/%s [id=%s]\n%s", n, tot, features_id, msg)

        def _group_features_by_attributes() -> tuple[
            list[FeaturesConfig],
            dict[FeaturesConfigKey, list[FeaturesConfig]],
        ]:
            cached: list[FeaturesConfig] = []
            groups: dict[FeaturesConfigKey, list[FeaturesConfig]] = defaultdict(list)
            tot = len(self._features_configs)
            for n, features_config in enumerate(self._features_configs, 1):
                L.info("Preprocessing features %s/%s [id=%s]", n, tot, features_config.id)
                if self.cache_manager.get_cached_features_checksums(features_config):
                    # cached, append the features config
                    cached.append(features_config)
                else:
                    # not cached, group the features_config by common keys
                    key = FeaturesConfigKey.from_config(features_config)
                    groups[key].append(features_config)
            return cached, groups

        def _process_cached_features(cached: list[FeaturesConfig]) -> None:
            for n, features_config in enumerate(cached, 1):
                query = None
                if self._repo.cache_manager.features_cache_needs_filter(features_config):
                    query = {SIMULATION_ID: self._repo.simulation_ids}
                df_dict = self.cache_manager.load_features(features_config=features_config)
                features = _calculate_cached(features_config, df_dict, query=query)
                _process_features(features_config, features)
                _log_features(features, n, len(cached), features_config.id)

        def _process_new_features(groups: dict[FeaturesConfigKey, list[FeaturesConfig]]) -> None:
            for num, (features_configs_key, features_configs_list) in enumerate(groups.items(), 1):
                L.info("Considering group: %s/%s, key: %s", num, len(groups), features_configs_key)
                for n, (features_config, features) in enumerate(
                    _calculate_new(self._repo, features_configs_key, features_configs_list), 1
                ):
                    _process_features(features_config, features)
                    _log_features(features, n, len(features_configs_list), features_config.id)

        with timed(L.info, "Step 1: grouping features by attributes"):
            features_configs_cached, features_configs_groups = _group_features_by_attributes()
        with timed(L.info, "Step 2: processing cached features"):
            _process_cached_features(features_configs_cached)
        with timed(L.info, "Step 3: processing new features"):
            _process_new_features(features_configs_groups)
        L.info("Features calculation completed")

    def apply_filter(self, repo: Repository) -> "FeaturesCollection":
        """Apply a filter based on the extracted simulations and return a new object."""
        return FilteredFeaturesCollection(parent=self, repo=repo)


class FilteredFeaturesCollection(FeaturesCollection):
    """FilteredFeaturesCollection class."""

    def __init__(self, parent: FeaturesCollection, repo: Repository) -> None:
        """Init from an existing FeaturesCollection filtered by the simulations ids in repo.

        Filtered dataframes are never written to disk.
        """
        super().__init__(
            features_configs=parent._features_configs,
            repo=repo,
            cache_manager=parent.cache_manager.to_readonly(),
        )
        dataframes = {name: features.df for name, features in parent._data.items()}
        query = {SIMULATION_ID: repo.simulation_ids}
        self._data = _dataframes_to_features(dataframes, config=None, cached=True, query=query)
        self._concatenated_features = self._clone_concatenated_features(
            parent._concatenated_features
        )

    def _clone_concatenated_features(
        self, concatenated_features: dict[str, ConcatenatedFeatures]
    ) -> dict[str, ConcatenatedFeatures]:
        """Return a dict of cloned concatenated_features."""
        return {name: cf.clone(parent=self) for name, cf in concatenated_features.items()}


def _dataframes_to_features(
    df_dict: dict[str, pd.DataFrame],
    config: Optional[FeaturesConfig],
    cached: bool,
    query: Optional[dict],
) -> dict[str, Feature]:
    """Wrap the given DataFrames with Feature objects, and return the resulting dict.

    The DataFrames are filtered by the simulation_ids contained in the internal repo.

    Args:
        df_dict: dict of DataFrames to be wrapped.
        config: config converted to dict and assigned to attrs["config"] for each DataFrame.
        cached: True if the data is loaded from the cache, False otherwise.
        query: optional query that can be used to filter the features dataframes.
    """
    result = {}
    for name, df in df_dict.items():
        result[name] = Feature.from_pandas(df, query=query, cached=cached)
        if config is not None:
            # make a copy of the config accessible from the features dataframe attrs
            result[name].df.attrs["config"] = config.dict()
    return result


def _calculate_cached(
    features_config: FeaturesConfig,
    df_dict: dict[str, pd.DataFrame],
    query: Optional[dict],
) -> dict[str, Feature]:
    """Load cached features from a dict of DataFrames."""
    return _dataframes_to_features(df_dict, config=features_config, cached=True, query=query)


def _calculate_new(
    repo: Repository,
    features_configs_key: FeaturesConfigKey,
    features_configs_list: list[FeaturesConfig],
) -> Iterator[tuple[FeaturesConfig, dict[str, Feature]]]:
    """Calculate new features and yield tuples."""
    results = _calculate_features(
        repo=repo,
        features_configs_key=features_configs_key,
        features_configs_list=features_configs_list,
    )
    assert len(features_configs_list) == len(results)
    for features_config, df_dict in zip(features_configs_list, results):
        features = _dataframes_to_features(
            df_dict, config=features_config, cached=False, query=None
        )
        yield features_config, features


def _user_func_wrapper(
    task_index: int,
    key: NamedTuple,
    neurons_df: pd.DataFrame,
    windows_df: pd.DataFrame,
    report_df: pd.DataFrame,
    repo: Repository,
    features_config: FeaturesConfig,
    temp_folder: Path,
) -> None:
    """Call the user function for the specified key, and save the resulting DataFrames.

    It should be executed in a subprocess.

    Args:
        task_index: incremental index.
        key: namedtuple specifying the filter.
        neurons_df: filtered neurons DataFrame.
        windows_df:  filtered windows DataFrame.
        report_df:  filtered report DataFrame.
        repo: Repository instance.
        features_config: features configuration.
        temp_folder: path to the shared memory (recommended) or temp directory.
    """
    L.info("Calculating features for %s", key)
    merged_df = neurons_df.merge(windows_df, how="left").merge(report_df, how="left")
    # The params dict is deepcopied because it could be modified in the user function.
    # It could happen even with multiprocessing, because joblib may process tasks in batch.
    func = import_by_string(features_config.function)
    features_dict = func(repo=repo, key=key, df=merged_df, params=deepcopy(features_config.params))
    # compatibility with features defined with type=single
    if features_config.type == "single":
        features_dict = {features_config.name: features_dict}
    # verify and process the result
    if not isinstance(features_dict, dict):
        raise ValueError("The user function must return a dict of dataframes")
    for feature_group, result_df in features_dict.items():
        if not isinstance(result_df, pd.DataFrame):
            raise ValueError(f"Expected a DataFrame, not {type(result_df).__name__}")
        # ignore the index if it's unnamed and with one level; this can be useful
        # for example when the returned DataFrame has a RangeIndex to be dropped
        drop = result_df.index.names == [None]
        if features_config.multi_index:
            result_df = result_df.etl.add_conditions(conditions=key._fields, values=key, drop=drop)
        else:
            result_df.reset_index(drop=drop, inplace=True)
            result_df.etl.insert_columns(loc=0, columns=key._fields, values=key)
        # the conversion to the desired dtype here is important to reduce memory usage and cpu time
        result_df = ensure_dtypes(result_df)
        output_dir = temp_folder / f"{feature_group}{features_config.suffix}"
        if not output_dir.is_dir():
            # the directory should be created in the first process
            output_dir.mkdir(parents=True, exist_ok=True)
        ParquetStore(output_dir).dump(result_df, name=f"{task_index:08d}")


def _merge_filter_func(
    task_index: int,
    key: NamedTuple,
    df_list: list[pd.DataFrame],
    temp_folder: Path,
    repo: Repository,
    features_configs_list: list[FeaturesConfig],
) -> None:
    """Executed in a subprocess, call the wrapper function for each features_config."""
    neurons_df, windows_df, report_df = df_list
    for features_config_index, features_config in enumerate(features_configs_list):
        _user_func_wrapper(
            task_index=task_index,
            key=key,
            neurons_df=neurons_df,
            windows_df=windows_df,
            report_df=report_df,
            repo=repo,
            features_config=features_config,
            temp_folder=temp_folder / str(features_config_index),
        )


def _filter_by_value(df: pd.DataFrame, key: str, value: Any) -> pd.DataFrame:
    """Filter the DataFrame only if the specified value is not None or empty."""
    return df.etl.q({key: value}) if value else df


@isolated
def _merge_filter_wrapper(
    temp_folder: Path,
    repo: Repository,
    features_configs_key: FeaturesConfigKey,
    features_configs_list: list[FeaturesConfig],
) -> None:
    """Execute merge_filter in an isolated subprocess.

    It's faster than running in the main process, for example the subprocess
    can take 167 seconds instead of 266 seconds (needing more investigation).

    With the current approach:

    - repo is pickled and passed to the subprocess
    - the needed dataframes are loaded from the cache in the subprocess
    - this is faster than passing big dataframes as parameters, that should be pickled
      in the main process and unpickled in the subprocess.
    """
    func = partial(
        _merge_filter_func,
        temp_folder=temp_folder,
        features_configs_list=features_configs_list,
        repo=repo,
    )
    merge_filter(
        df_list=[
            _filter_by_value(
                repo.neurons.df,
                key="neuron_class",
                value=features_configs_key.neuron_classes,
            ),
            _filter_by_value(
                repo.windows.df,
                key="window",
                value=features_configs_key.windows,
            ),
            repo.report.df,
        ],
        groupby=features_configs_key.groupby,
        func=func,
    )


def _concatenate_all(temp_folder: Path) -> list[dict[str, pd.DataFrame]]:
    """Concatenate all the dataframes having the same feature_group label.

    The DataFrames are loaded from parquet files contained in a folder structure like:

    /temp_folder
    ├── 0
    │   ├── by_gid
    │   │   └── *.parquet
    │   ├── by_gid_and_trial
    │   │   └── *.parquet
    │   ├── by_neuron_class
    │   │   └── *.parquet
    │   └── by_neuron_class_and_trial
    │       └── *.parquet
    ├── 1
    │   └── other_features
    │       └── *.parquet
    ...
    └── n
        └── other_features
            └── *.parquet

    where each numerical folder corresponds to one entry in features_configs_list,
    and contains one subfolder for each DataFrame to be created.

    Returns:
        list of DataFrames obtained by the concatenation of the partial DataFrames.
    """
    result = []
    # the numerical subdirectories must be processed in ascending order
    for index_path in sorted(temp_folder.iterdir(), key=lambda p: int(p.name)):
        d = {}
        # the dataframes subdirectories can be processed in any order
        for features_path in sorted(index_path.iterdir()):
            df = ParquetStore(features_path).load()
            df = ensure_dtypes(df)
            d[features_path.name] = df
        result.append(d)
    return result


def _calculate_features(
    repo: Repository,
    features_configs_key: FeaturesConfigKey,
    features_configs_list: list[FeaturesConfig],
) -> list[dict[str, pd.DataFrame]]:
    """Calculate features in parallel for the given repository as a dict of DataFrames.

    Args:
        repo: repository containing spikes.
        features_configs_key: common key of the features configurations.
        features_configs_list: list of features configurations.

    Returns:
        list of dicts of features DataFrames, one item for each features_config.
    """
    with tempfile.TemporaryDirectory(prefix="blueetl_", dir=get_shmdir()) as _temp_folder:
        with timed(L.info, "Executing merge_filter"):
            _merge_filter_wrapper(
                temp_folder=Path(_temp_folder),
                repo=repo,
                features_configs_key=features_configs_key,
                features_configs_list=features_configs_list,
            )
        with timed(L.info, "Executing concatenation"):
            return _concatenate_all(temp_folder=Path(_temp_folder))



import logging
import sys
from pathlib import Path
from unittest.mock import Mock, patch

import pytest
from click.testing import CliRunner

from blueetl.apps import run as test_module


@pytest.mark.parametrize("skip_features_cache", [True, False, None])
@pytest.mark.parametrize("readonly_cache", [True, False, None])
@pytest.mark.parametrize("clear_cache", [True, False, None])
@pytest.mark.parametrize("show", [True, False])
@pytest.mark.parametrize("calculate", [True, False])
@pytest.mark.parametrize("extract", [True, False])
@patch(test_module.__name__ + ".run_from_file")
def test_run(
    mock_run_from_file,
    tmp_path,
    extract,
    calculate,
    show,
    clear_cache,
    readonly_cache,
    skip_features_cache,
):
    options_dict = {
        "extract": extract,
        "calculate": calculate,
        "show": show,
        "clear-cache": clear_cache,
        "readonly-cache": readonly_cache,
        "skip-features-cache": skip_features_cache,
    }
    options = [f"--{k}" if v else f"--no-{k}" for k, v in options_dict.items() if v is not None]
    analysis_config_file = "config.yaml"
    runner = CliRunner()

    with runner.isolated_filesystem(temp_dir=tmp_path) as td:
        Path(analysis_config_file).write_text("---")
        result = runner.invoke(test_module.run, [analysis_config_file, "-vv", *options])

    assert result.output == ""
    assert result.exit_code == 0
    mock_run_from_file.assert_called_once_with(
        analysis_config_file=analysis_config_file,
        seed=0,
        extract=extract,
        calculate=calculate,
        show=show,
        clear_cache=clear_cache,
        readonly_cache=readonly_cache,
        skip_features_cache=skip_features_cache,
        loglevel=logging.DEBUG,
    )


@patch.dict(sys.modules, {"IPython": Mock()})
@patch(test_module.__name__ + ".run_from_file")
def test_run_interactive_success(mock_run_from_file, tmp_path):
    analysis_config_file = "config.yaml"
    runner = CliRunner()

    with runner.isolated_filesystem(temp_dir=tmp_path) as td:
        Path(analysis_config_file).write_text("---")
        result = runner.invoke(test_module.run, [analysis_config_file, "--interactive"])

    assert result.output == ""
    assert result.exit_code == 0
    assert mock_run_from_file.call_count == 1
    assert sys.modules["IPython"].embed.call_count == 1


@patch.dict(sys.modules, {"IPython": None})
@patch(test_module.__name__ + ".run_from_file")
def test_run_interactive_failure(mock_run_from_file, tmp_path):
    analysis_config_file = "config.yaml"
    runner = CliRunner()

    with runner.isolated_filesystem(temp_dir=tmp_path) as td:
        Path(analysis_config_file).write_text("---")
        result = runner.invoke(test_module.run, [analysis_config_file, "--interactive"])

    assert result.output.strip() == "You need to install IPython to start an interactive session."
    assert result.exit_code == 1
    assert mock_run_from_file.call_count == 1



"""Features extractor."""

import logging

import pandas as pd

from blueetl.extract.base import BaseExtractor

L = logging.getLogger(__name__)


class Feature(BaseExtractor):
    """Features extractor class."""

    @classmethod
    def _validate(cls, df: pd.DataFrame) -> None:
        # no validation is needed for features
        pass



"""Generic Report extractor."""

import logging
import tempfile
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Callable, NamedTuple, Optional, TypeVar

import pandas as pd
from blueetl_core.utils import smart_concat

from blueetl.adapters.simulation import SimulationAdapter as Simulation
from blueetl.constants import CIRCUIT_ID, GID, NEURON_CLASS, POPULATION, SIMULATION, SIMULATION_ID
from blueetl.extract.base import BaseExtractor
from blueetl.extract.neuron_classes import NeuronClasses
from blueetl.extract.neurons import Neurons
from blueetl.extract.simulations import Simulations
from blueetl.extract.windows import Windows
from blueetl.parallel import merge_filter
from blueetl.store.parquet import ParquetStore
from blueetl.utils import ensure_dtypes, get_shmdir, timed

L = logging.getLogger(__name__)
ReportExtractorT = TypeVar("ReportExtractorT", bound="ReportExtractor")


@dataclass
class WindowSlice:
    """Window slice attributes."""

    t_start: float
    t_stop: float
    t_step: Optional[float]
    offset: float
    name: str
    trial: int


class ReportExtractor(BaseExtractor, metaclass=ABCMeta):
    """Report extractor class."""

    @staticmethod
    def calculate_window_slice(rec) -> WindowSlice:
        """Calculate and return the window slice attributes."""
        # increment t_start and t_stop because they are relative to offset
        t_start = rec.offset + rec.t_start
        t_stop = rec.offset + rec.t_stop
        t_step = rec.t_step or None
        return WindowSlice(
            t_start=t_start,
            t_stop=t_stop,
            t_step=t_step,
            offset=rec.offset,
            name=rec.window,
            trial=rec.trial,
        )

    @classmethod
    @abstractmethod
    def _load_values(
        cls,
        simulation: Simulation,
        population: Optional[str],
        gids,
        windows_df: pd.DataFrame,
        name: str,
    ) -> pd.DataFrame:
        """Return a DataFrame for the given simulation, population, gids, and windows.

        Args:
            simulation: simulation containing the report.
            population: node population name.
            gids: array of gids to be selected.
            windows_df: windows dataframe.
            name: name of the report in the simulation configuration.

        Returns:
            pd.DataFrame: dataframe with the needed columns.
        """

    @classmethod
    def from_simulations(
        cls: type[ReportExtractorT],
        simulations: Simulations,
        neurons: Neurons,
        windows: Windows,
        neuron_classes: NeuronClasses,
        name: str,
    ) -> ReportExtractorT:
        """Return a new instance from the given simulations, neurons, and windows.

        Args:
            simulations: Simulations extractor.
            neurons: Neurons extractor.
            windows: Windows extractor.
            neuron_classes: NeuronClasses extractor.
            name: name of the report in the simulation configuration.

        Returns:
            New instance.
        """
        with tempfile.TemporaryDirectory(prefix="blueetl_", dir=get_shmdir()) as _temp_folder:
            with timed(L.info, "Executing merge_filter "):
                func = partial(
                    _merge_filter_func,
                    temp_folder=Path(_temp_folder),
                    name=name,
                    neuron_classes_df=neuron_classes.df,
                    dataframe_builder=cls._load_values,
                )
                merge_filter(
                    df_list=[simulations.df, neurons.df, windows.df],
                    groupby=[SIMULATION_ID, CIRCUIT_ID],
                    func=func,
                )
            with timed(L.info, "Executing concatenation"):
                df = ParquetStore(Path(_temp_folder)).load()
                df = ensure_dtypes(df)
            return cls(df, cached=False, filtered=False)


def _merge_filter_func(
    task_index: int,
    key: NamedTuple,
    df_list: list[pd.DataFrame],
    temp_folder: Path,
    name: str,
    neuron_classes_df: pd.DataFrame,
    dataframe_builder: Callable[..., pd.DataFrame],
) -> None:
    """Executed in a subprocess, write a partial DataFrame to temp_folder."""
    # pylint: disable=too-many-locals
    simulations_df, neurons_df, windows_df = df_list
    simulation_id, simulation = simulations_df.etl.one()[[SIMULATION_ID, SIMULATION]]
    assert simulation_id == key.simulation_id  # type: ignore[attr-defined]
    df_list = []
    for inner_key, df in neurons_df.etl.groupby_iter([CIRCUIT_ID, NEURON_CLASS]):
        population = neuron_classes_df.etl.one(
            circuit_id=inner_key.circuit_id, neuron_class=inner_key.neuron_class
        )[POPULATION]
        result_df = dataframe_builder(
            simulation=simulation,
            population=population,
            gids=df[GID],
            windows_df=windows_df,
            name=name,
        )
        result_df[[SIMULATION_ID, *inner_key._fields]] = [simulation_id, *inner_key]
        df_list.append(result_df)
    result_df = smart_concat(df_list, ignore_index=True)
    # the conversion to the desired dtype here is important to reduce memory usage and cpu time
    result_df = ensure_dtypes(result_df)
    ParquetStore(temp_folder).dump(result_df, name=f"{task_index:08d}")



"""Simulation campaign utils.

Originally based on:
https://bbpgitlab.epfl.ch/nse/bbp-workflow/-/blob/2032ffd6/bbp_workflow/simulation/util.py
"""

from collections.abc import Iterator
from typing import Any, Optional, Union

import xarray as xr

from blueetl.campaign.config import SimulationCampaign

DataArrayOrDict = Union[dict, xr.DataArray]


def from_xarray(config: DataArrayOrDict) -> SimulationCampaign:
    """Return a SimulationCampaign from xarray.DataArray or dict."""
    if isinstance(config, dict):
        return SimulationCampaign.from_xarray_dict(config)
    else:
        return SimulationCampaign.from_xarray(config)


def campaign_sims(
    config: DataArrayOrDict, include_empty: bool = False
) -> Iterator[tuple[dict[str, Any], Optional[str]]]:
    """Iterate simulations from the campaign.

    Campaign sim can be empty if it was filtered out by `coords_filter_func`.

    Args:
        config (xarray.DataArray|dict): Simulation campaign configuration.
        include_empty (bool): If true, all sims are included(even empty ones).

    Yields:
        tuple(dict, str): Simulation conditions and full path to individual simulation.
                          Path can be `None` if sim folder is empty.
    """
    config = from_xarray(config)
    for sim in config:
        if include_empty or not sim.empty:
            yield sim.conditions, sim.path or None


def _campaign_sims_with_indices(config: DataArrayOrDict) -> Iterator[tuple[dict, dict, str]]:
    """Iterate all simulations from the non-coupled coords campaign with indices.

    Args:
        config (xarray.DataArray|dict): Simulation campaign configuration.

    Yields:
        tuple(dict, dict, str): Simulation conditions with locs/indices and full path to
                                individual simulations.
    """
    config = from_xarray(config)
    if config.is_coupled():
        raise ValueError("Please provide non-coupled coords sim campaign!")
    # number of distinct values for each condition
    totals = config.conditions.nunique().to_dict()
    # indices for each condition and simulation
    idxs = {cond: series.cat.codes for cond, series in config.conditions.astype("category").items()}
    for sim in config:
        indices = {"idx": sim.index}
        for cond in sim.conditions:
            indices[f"{cond}_idx"] = idxs[cond].at[sim.index]
            indices[f"{cond}_total"] = totals[cond]
        yield sim.conditions, indices, sim.path


def _campaign_sim_indices(config: DataArrayOrDict) -> list[int]:
    """Non-empty simulation indices of the campaign."""
    config = from_xarray(config)
    return [sim.index for sim in config if sim.path]


def _campaign_sim_index_to_coords(config: DataArrayOrDict) -> dict[int, dict[str, Any]]:
    """Index to coords dict of the campaign sims."""
    config = from_xarray(config)
    return {sim.index: sim.conditions for sim in config if sim.path}



import pickle
import shutil
from pathlib import Path
from unittest.mock import patch

import pytest

from blueetl import analysis as test_module
from blueetl.config.analysis_model import MultiAnalysisConfig
from tests.unit.utils import TEST_DATA_PATH


def _prepare_env(path):
    """Copy test data to the given path, and return the path to the analysis config."""
    (path / "circuit").mkdir()
    (path / "analysis").mkdir()
    shutil.copytree(TEST_DATA_PATH / "circuit" / "sonata", path / "circuit" / "sonata")
    shutil.copytree(TEST_DATA_PATH / "simulation_campaign", path / "simulation_campaign")
    return Path(
        shutil.copy(
            TEST_DATA_PATH / "analysis" / "analysis_config_01_relative.yaml",
            path / "analysis",
        )
    )


@pytest.mark.parametrize("skip_features_cache", [True, False, None])
@pytest.mark.parametrize("readonly_cache", [True, False, None])
@pytest.mark.parametrize("clear_cache", [True, False, None])
@pytest.mark.parametrize("show", [True, False])
@pytest.mark.parametrize("calculate", [True, False])
@pytest.mark.parametrize("extract", [True, False])
@patch.object(test_module.MultiAnalyzer, "from_file")
def test_run_from_file(
    from_file,
    tmp_path,
    extract,
    calculate,
    show,
    clear_cache,
    readonly_cache,
    skip_features_cache,
):
    analysis_config_file = tmp_path / "config.yaml"
    analysis_config_file.write_text("---")

    instance = test_module.run_from_file(
        analysis_config_file=analysis_config_file,
        extract=extract,
        calculate=calculate,
        show=show,
        clear_cache=clear_cache,
        readonly_cache=readonly_cache,
        skip_features_cache=skip_features_cache,
    )

    from_file.assert_called_once_with(
        analysis_config_file,
        extra_params={
            "clear_cache": clear_cache,
            "readonly_cache": readonly_cache,
            "skip_features_cache": skip_features_cache,
        },
    )
    assert instance.extract_repo.call_count == int(extract)
    assert instance.calculate_features.call_count == int(calculate)
    assert instance.show.call_count == int(show)


def test_multi_analyzer_from_file(tmp_path):
    path = _prepare_env(tmp_path)
    with test_module.MultiAnalyzer.from_file(path, extra_params={}) as ma:
        assert isinstance(ma, test_module.MultiAnalyzer)
        assert isinstance(ma.global_config, MultiAnalysisConfig)
        assert isinstance(ma.analyzers, dict)
        assert ma.names == ["spikes"]
        assert isinstance(ma.spikes, test_module.Analyzer)

        filtered = ma.apply_filter()
        assert isinstance(filtered, test_module.MultiAnalyzer)
        assert filtered is ma

        filtered = ma.apply_filter(simulations_filter={"seed": 334630})
        assert isinstance(filtered, test_module.MultiAnalyzer)
        assert filtered is not ma

        dumped = pickle.dumps(ma)
        loaded = pickle.loads(dumped)

        assert isinstance(loaded, test_module.MultiAnalyzer)



from unittest.mock import Mock, PropertyMock

import pandas as pd
from pandas.testing import assert_frame_equal

from blueetl.config.analysis_model import TrialStepsConfig, WindowConfig
from blueetl.constants import (
    CIRCUIT,
    CIRCUIT_ID,
    DURATION,
    OFFSET,
    SIMULATION,
    SIMULATION_ID,
    T_START,
    T_STEP,
    T_STOP,
    TRIAL,
    WINDOW,
    WINDOW_TYPE,
)
from blueetl.extract import windows as test_module
from blueetl.resolver import AttrResolver
from blueetl.utils import ensure_dtypes
from tests.unit.utils import TEST_NODE_SETS_FILE_EXTRA


def _myfunc1(spikes, params):
    """Calculate and return the cortical onset from spikes"""
    return 250


def _myfunc2(spikes, params):
    """Calculate and return the cortical onset from spikes"""
    return 750


def test_windows_from_simulations(mock_simulation, mock_circuit):
    mock_simulations_df = PropertyMock(
        return_value=pd.DataFrame(
            [
                {
                    SIMULATION_ID: 0,
                    CIRCUIT_ID: 0,
                    SIMULATION: mock_simulation,
                    CIRCUIT: mock_circuit,
                },
            ]
        )
    )
    mock_simulations = Mock()
    type(mock_simulations).df = mock_simulations_df

    root = Mock()
    root.soma.repo.windows.df = pd.DataFrame(
        {
            SIMULATION_ID: 0,
            CIRCUIT_ID: 0,
            WINDOW: "w9",
            TRIAL: 0,
            OFFSET: 1000,
            T_START: 0,
            T_STOP: 900,
            T_STEP: 0,
            DURATION: 900,
            WINDOW_TYPE: "custom_type_9",
        },
        index=[0],
    )
    root.soma.extraction = root.soma.repo
    resolver = AttrResolver(root)

    windows_config = {
        "w1": WindowConfig(bounds=[5000, 6000]),
        "w2": WindowConfig(
            **{
                "bounds": [0, 100],
                "initial_offset": 1000,
                "window_type": "custom_type_1",
                "n_trials": 2,
                "trial_steps_value": 100,
            }
        ),
        "w3": WindowConfig(
            **{
                "bounds": [0, 200],
                "initial_offset": 2000,
                "window_type": "custom_type_2",
                "trial_steps_list": [0, 150],
            }
        ),
        "w4": "soma.extraction.windows.w9#checksum",
        "w5": WindowConfig(
            **{
                "bounds": [0, 200],
                "initial_offset": 2000,
                "window_type": "custom_type_3",
                "trial_steps_label": "ts1",
            }
        ),
        "w6": WindowConfig(
            **{
                "bounds": [0, 200],
                "initial_offset": 2000,
                "window_type": "custom_type_4",
                "trial_steps_list": [50, 150],
                "trial_steps_label": "ts2",
            }
        ),
    }
    trial_steps_config = {
        "ts1": TrialStepsConfig(
            function=f"{__name__}._myfunc1",
            bounds=[-50, 150],
        ),
        "ts2": TrialStepsConfig(
            function=f"{__name__}._myfunc2",
            bounds=[-20, 200],
            node_set="ExtraLayer2",
            node_sets_file=TEST_NODE_SETS_FILE_EXTRA,
        ),
    }

    result = test_module.Windows.from_simulations(
        simulations=mock_simulations,
        windows_config=windows_config,
        trial_steps_config=trial_steps_config,
        resolver=resolver,
    )

    expected_df = pd.DataFrame(
        [
            {
                SIMULATION_ID: 0,
                CIRCUIT_ID: 0,
                WINDOW: "w1",
                TRIAL: 0,
                OFFSET: 0,
                T_START: 5000,
                T_STOP: 6000,
                T_STEP: 0,
                DURATION: 1000,
                WINDOW_TYPE: "",
            },
            {
                SIMULATION_ID: 0,
                CIRCUIT_ID: 0,
                WINDOW: "w2",
                TRIAL: 0,
                OFFSET: 1000,
                T_START: 0,
                T_STOP: 100,
                T_STEP: 0,
                DURATION: 100,
                WINDOW_TYPE: "custom_type_1",
            },
            {
                SIMULATION_ID: 0,
                CIRCUIT_ID: 0,
                WINDOW: "w2",
                TRIAL: 1,
                OFFSET: 1100,
                T_START: 0,
                T_STOP: 100,
                T_STEP: 0,
                DURATION: 100,
                WINDOW_TYPE: "custom_type_1",
            },
            {
                SIMULATION_ID: 0,
                CIRCUIT_ID: 0,
                WINDOW: "w3",
                TRIAL: 0,
                OFFSET: 2000,
                T_START: 0,
                T_STOP: 200,
                T_STEP: 0,
                DURATION: 200,
                WINDOW_TYPE: "custom_type_2",
            },
            {
                SIMULATION_ID: 0,
                CIRCUIT_ID: 0,
                WINDOW: "w3",
                TRIAL: 1,
                OFFSET: 2150,
                T_START: 0,
                T_STOP: 200,
                T_STEP: 0,
                DURATION: 200,
                WINDOW_TYPE: "custom_type_2",
            },
            {
                SIMULATION_ID: 0,
                CIRCUIT_ID: 0,
                WINDOW: "w4",
                TRIAL: 0,
                OFFSET: 1000,
                T_START: 0,
                T_STOP: 900,
                T_STEP: 0,
                DURATION: 900,
                WINDOW_TYPE: "custom_type_9",
            },
            {
                SIMULATION_ID: 0,
                CIRCUIT_ID: 0,
                WINDOW: "w5",
                TRIAL: 0,
                OFFSET: 2250,
                T_START: 0,
                T_STOP: 200,
                T_STEP: 0,
                DURATION: 200,
                WINDOW_TYPE: "custom_type_3",
            },
            {
                SIMULATION_ID: 0,
                CIRCUIT_ID: 0,
                WINDOW: "w6",
                TRIAL: 0,
                OFFSET: 2800,
                T_START: 0,
                T_STOP: 200,
                T_STEP: 0,
                DURATION: 200,
                WINDOW_TYPE: "custom_type_4",
            },
            {
                SIMULATION_ID: 0,
                CIRCUIT_ID: 0,
                WINDOW: "w6",
                TRIAL: 1,
                OFFSET: 2900,
                T_START: 0,
                T_STOP: 200,
                T_STEP: 0,
                DURATION: 200,
                WINDOW_TYPE: "custom_type_4",
            },
        ]
    )
    expected_df = ensure_dtypes(expected_df)
    assert isinstance(result, test_module.Windows)
    assert_frame_equal(result.df, expected_df)
    assert mock_simulations_df.call_count == 1



"""Bluepy simulation implementation."""

from collections import UserDict
from collections.abc import Mapping
from functools import cached_property
from typing import Optional, Union

import pandas as pd
from bluepy import Simulation
from bluepy.exceptions import BluePyError
from bluepy.impl.compartment_report import CompartmentReport, SomaReport
from bluepy.impl.spike_report import SpikeReport
from bluepy.simulation import PathHelpers

from blueetl.adapters.impl.bluepy.circuit import CircuitImpl
from blueetl.adapters.interfaces.circuit import CircuitInterface
from blueetl.adapters.interfaces.simulation import (
    PopulationReportInterface,
    PopulationSpikesReportInterface,
    SimulationInterface,
)


class PopulationSpikesReportImpl(PopulationSpikesReportInterface):
    """Bluepy spikes report implementation."""

    def __init__(self, spikes: SpikeReport) -> None:
        """Init the spikes report with the given spikes."""
        self._spikes = spikes

    def get(self, group=None, t_start=None, t_stop=None) -> pd.Series:
        """Return the spikes report for the specified group and interval.

        Only `group` as array of numeric ids is supported.
        """
        return self._spikes.get(gids=group, t_start=t_start, t_end=t_stop)


class PopulationReportImpl(PopulationReportInterface):
    """Bluepy report implementation."""

    def __init__(self, report: Union[SomaReport, CompartmentReport]) -> None:
        """Init the population report with the given report."""
        self._report = report

    def get(self, group=None, t_start=None, t_stop=None, t_step=None) -> pd.DataFrame:
        """Return the report for the specified group and interval.

        Only `group` as array of numeric ids is supported.
        """
        return self._report.get(gids=group, t_start=t_start, t_end=t_stop, t_step=t_step)


class ReportCollection(UserDict):
    """Collection of reports as: name -> population -> report."""

    def __init__(self, simulation: Simulation) -> None:
        """Init the report collection with the specified simulation."""
        super().__init__()
        self._simulation = simulation

    def __getitem__(self, name) -> Mapping[Optional[str], PopulationReportInterface]:
        """Return the report for the specified name, wrapped in a dict as: population -> report.

        The population name in the returned dict is always None, because undefined in the config.
        """
        if name not in self.data:
            self.data[name] = {None: PopulationReportImpl(self._simulation.report(name))}
        return self.data[name]


class SimulationImpl(SimulationInterface[Simulation]):
    """Bluepy simulation implementation."""

    def is_complete(self) -> bool:
        """Return True if the simulation is complete, False otherwise."""
        try:
            # check the existence of spikes without loading them, because it can be slow
            PathHelpers.spike_report_path(self._simulation.config)
            return True
        except BluePyError:
            return False

    @cached_property
    def circuit(self) -> CircuitInterface:
        """Return the circuit used for the simulation."""
        return CircuitImpl(self._simulation.circuit)

    @cached_property
    def spikes(self) -> Mapping[Optional[str], PopulationSpikesReportInterface]:
        """Return the spikes report as a dict: population -> report.

        The population name in the returned dict is always None, because undefined in the config.
        """
        return {None: PopulationSpikesReportImpl(self._simulation.spikes)}

    @cached_property
    def reports(self) -> Mapping[str, Mapping[Optional[str], PopulationReportInterface]]:
        """Return the reports as a dict: name -> population -> report.

        The population name in the returned dict is always None, because undefined in the config.
        """
        return ReportCollection(simulation=self._simulation)



import pandas as pd
import pytest
from pandas.testing import assert_frame_equal
from pyarrow import ArrowNotImplementedError

from blueetl.store import parquet as test_module


@pytest.mark.parametrize(
    "df",
    [
        "storable_df_with_unnamed_index",
        "storable_df_with_named_index",
        "storable_df_with_named_multiindex",
        "storable_df_with_unnamed_multiindex",
    ],
)
def test_dump_load_roundtrip(tmp_path, df, lazy_fixture):
    df = lazy_fixture(df)
    name = "myname"

    store = test_module.ParquetStore(tmp_path)
    store.dump(df, name)
    result = store.load(name)

    assert_frame_equal(result, df)


@pytest.mark.parametrize(
    "df",
    [
        "storable_df_with_unnamed_index",
        "storable_df_with_named_index",
        "storable_df_with_named_multiindex",
        "storable_df_with_unnamed_multiindex",
    ],
)
def test_dump_load_roundtrip_with_inferred_schema(tmp_path, df, lazy_fixture):
    df = lazy_fixture(df)
    df1 = df.copy()
    df2 = df.copy()
    df1 = df1.etl.add_conditions(conditions=["extra_level"], values=[1])
    df2 = df2.etl.add_conditions(conditions=["extra_level"], values=[2])

    df1.insert(loc=0, column="extra_columns", value=[[] for _ in range(len(df1))])
    df2.insert(loc=0, column="extra_columns", value=[[float(i)] for i in range(len(df2))])

    df = pd.concat([df1, df2])

    store = test_module.ParquetStore(tmp_path)
    store.dump(df1, name="01")
    store.dump(df2, name="02")
    result = store.load()

    assert_frame_equal(result, df)

    # in pyarrow 15.0.0, the empty lists in the first DataFrame would cause:
    #
    #   pyarrow.lib.ArrowNotImplementedError:
    #     Unsupported cast from double to null using function cast_null
    #
    # if the schema is not inferred correctly from both the files
    with pytest.raises(ArrowNotImplementedError):
        store.load(schema=None)



"""Neurons extractor."""

import logging
from pathlib import Path
from typing import Optional

import numpy as np
import pandas as pd

from blueetl.adapters.circuit import CircuitAdapter as Circuit
from blueetl.adapters.node_sets import NodeSetsAdapter as NodeSets
from blueetl.config.analysis_model import NeuronClassConfig
from blueetl.constants import CIRCUIT, CIRCUIT_ID, GID, NEURON_CLASS, NEURON_CLASS_INDEX
from blueetl.extract.base import BaseExtractor
from blueetl.extract.simulations import Simulations
from blueetl.utils import ensure_list, timed

L = logging.getLogger(__name__)

# cached node_ids for each (population, node_set, node_sets_file)
CellsCache = dict[tuple[Optional[str], Optional[str], Optional[Path]], pd.DataFrame]


def _get_property_names(neuron_classes: dict[str, NeuronClassConfig]) -> list[str]:
    """Return the list of properties to be retrieved from the cells DataFrame."""
    properties_set = set()
    for conf in neuron_classes.values():
        for query_dict in ensure_list(conf.query):
            properties_set.update(query_dict)
    return sorted(properties_set)


def _load_cells(
    circuit: Circuit,
    property_names: list[str],
    cells_cache: CellsCache,
    population: Optional[str],
    node_set: Optional[str],
    node_sets_file: Optional[Path],
) -> pd.DataFrame:
    """Load and return the cells for the given population and node_set.

    Data are retrieved from the circuit or from the cache.

    If node_set is None or empty string, all the cells of the population are loaded.
    """
    node_set = node_set or None
    key = (population, node_set, node_sets_file)
    if key not in cells_cache:
        msg = f"Loading nodes using {population=}, {node_set=}, {node_sets_file=}"
        with timed(L.info, msg):
            if node_set and node_sets_file:
                node_sets = NodeSets.from_file(circuit.node_sets_file)
                node_sets |= NodeSets.from_file(node_sets_file)
                node_set = node_sets.instance[node_set]
            _cells = circuit.nodes[population].get(group=node_set, properties=property_names)
            cells_cache[key] = _cells
    return cells_cache[key]


def _filter_gids_by_neuron_class(
    circuit: Circuit,
    property_names: list[str],
    cells_cache: CellsCache,
    name: str,
    config: NeuronClassConfig,
) -> np.ndarray:
    """Return the array of node_ids filtered by neuron class."""
    cells = _load_cells(
        circuit=circuit,
        property_names=property_names,
        cells_cache=cells_cache,
        population=config.population,
        node_set=config.node_set,
        node_sets_file=config.node_sets_file,
    )
    gids = cells.etl.q(config.query).index.to_numpy()
    if config.node_id is not None:
        gids = np.intersect1d(gids, config.node_id)
    neuron_count = len(gids)
    if config.limit and neuron_count > config.limit:
        gids = np.random.choice(gids, size=config.limit, replace=False)
    gids.sort()
    L.info(
        "Selected gids for %s: %s/%s (limit=%s, population=%s, node_set=%s)",
        name,
        len(gids),
        neuron_count,
        config.limit,
        config.population,
        config.node_set,
    )
    L.debug("Configured query: %s", config.query)
    L.debug("Configured node_id: %s", config.node_id)
    return gids


class Neurons(BaseExtractor):
    """Neurons extractor class."""

    COLUMNS = [CIRCUIT_ID, NEURON_CLASS, GID, NEURON_CLASS_INDEX]

    def __init__(self, df: pd.DataFrame, cached: bool, filtered: bool) -> None:
        """Initialize the extractor."""
        super().__init__(df, cached=cached, filtered=filtered)
        # ensure that the neurons are sorted
        self._df: pd.DataFrame = self._df.sort_values(self.COLUMNS, ignore_index=True)

    @staticmethod
    def _get_gids(
        circuit: Circuit, neuron_classes: dict[str, NeuronClassConfig]
    ) -> dict[str, np.ndarray]:
        """Return a dict containing name: node_ids for each neuron class."""
        cells_cache: CellsCache = {}
        property_names = _get_property_names(neuron_classes=neuron_classes)
        return {
            name: _filter_gids_by_neuron_class(circuit, property_names, cells_cache, name, config)
            for name, config in neuron_classes.items()
        }

    @classmethod
    def from_simulations(
        cls, simulations: Simulations, neuron_classes: dict[str, NeuronClassConfig]
    ) -> "Neurons":
        """Return a new Neurons instance from the given simulations and configuration.

        Args:
            simulations: Simulations extractor.
            neuron_classes: configuration dict of neuron classes to be extracted.

        Returns:
            Neurons: new instance.
        """
        grouped = simulations.df.groupby([CIRCUIT_ID])[CIRCUIT].first()
        records: list[tuple[int, str, int, int]] = []
        for circuit_id, circuit in grouped.items():
            gids_by_class = cls._get_gids(circuit, neuron_classes=neuron_classes)
            records.extend(
                (circuit_id, neuron_class, gid, neuron_class_index)
                for neuron_class, gids in gids_by_class.items()
                for neuron_class_index, gid in enumerate(gids)
            )
        df = pd.DataFrame.from_records(records, columns=cls.COLUMNS)
        return cls(df, cached=False, filtered=False)

    def count_by_neuron_class(self, observed: bool = True) -> pd.Series:
        """Return the number of gids for each circuit and neuron class.

        Args:
            observed: If True: only show observed values for categorical groupers.
                If False: show all values for categorical groupers.
        """
        return self.df.groupby([CIRCUIT_ID, NEURON_CLASS], observed=observed)[GID].count()



import os
import shutil
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from blueetl.analysis import MultiAnalyzer
from blueetl.constants import CIRCUIT, SIMULATION
from blueetl.extract.feature import Feature
from blueetl.features import ConcatenatedFeatures
from tests.functional.utils import assertion_error_message


def _get_subattr(obj, attrs):
    for attr in attrs:
        obj = getattr(obj, attr)
    return obj


def _load_df(path):
    return pd.read_parquet(path)


def _dump_df(df, path):
    df.to_parquet(path)


def _dump_all(a, path):
    # used only when test cases are added or modified
    path = Path(path).resolve()
    a.extract_repo()
    a.calculate_features()
    for container_name in "repo", "features":
        container = getattr(a, container_name)
        for name in getattr(container, "names"):
            obj = getattr(container, name)
            if not isinstance(obj, ConcatenatedFeatures):
                # do not dump ConcatenatedFeatures
                df = obj.df
                df = df[[col for col in df.columns if col not in [SIMULATION, CIRCUIT]]]
                filepath = path / container_name / f"{name}.parquet"
                filepath.parent.mkdir(parents=True, exist_ok=True)
                _dump_df(df, filepath)


def _dump_all_multi(ma, path):
    # used only when test cases are added or modified
    path = Path(path).resolve()
    ma.extract_repo()
    ma.calculate_features()
    for name in ma.names:
        partial_path = path / name
        a = getattr(ma, name)
        _dump_all(a, partial_path)


def _test_repo(a, path):
    a.extract_repo()
    path = path / "repo"
    assert len(a.repo.names) > 0
    assert sorted(a.repo.names) == sorted(p.stem for p in path.glob("*.parquet"))
    for name in a.repo.names:
        expected_df = _load_df(path / f"{name}.parquet")
        df = _get_subattr(a, ["repo", name, "df"])
        df = df[[col for col in df.columns if col not in [SIMULATION, CIRCUIT]]]
        with assertion_error_message(f"Difference in repo {name!r}"):
            # check_index_type=False needed to avoid Attribute "inferred_type" are different
            assert_frame_equal(df, expected_df, check_index_type=len(expected_df) != 0)


def _test_features(a, path):
    a.calculate_features()
    path = path / "features"
    feature_names = []
    for name in a.features.names:
        obj = getattr(a.features, name)
        if isinstance(obj, Feature):
            feature_names.append(name)
            actual_df = obj.df
            expected_df = _load_df(path / f"{name}.parquet")
            with assertion_error_message(f"Difference in features {name!r}"):
                # in case of error: Attribute "inferred_type" are different
                # consider setting check_index_type=False with empty DataFrames
                assert_frame_equal(actual_df, expected_df)
        elif isinstance(obj, ConcatenatedFeatures):
            assert isinstance(obj.df, pd.DataFrame)
            assert isinstance(obj.params, pd.DataFrame)
            assert isinstance(obj.aliases, pd.DataFrame)
        else:
            raise TypeError(f"Invalid class for feature {name}: {obj.__class__.__name__}")
    assert sorted(feature_names) == sorted(p.stem for p in path.glob("*.parquet"))


def _test_repo_multi(ma, path):
    ma.extract_repo()
    for name in ma.names:
        partial_path = path / name
        a = getattr(ma, name)
        _test_repo(a, partial_path)


def _test_features_multi(ma, path):
    ma.calculate_features()
    for name in ma.names:
        partial_path = path / name
        a = getattr(ma, name)
        _test_features(a, partial_path)


def _test_filter_in_memory(ma, path):
    ma2 = ma.apply_filter()
    if not ma.global_config.simulations_filter_in_memory:
        assert ma2 is ma
    else:
        assert ma2 is not ma
        # test that the new DataFrames have been filtered
        _test_repo_multi(ma2, path / "_filtered")
        _test_features_multi(ma2, path / "_filtered")
        # test that the original DataFrames are unchanged
        _test_repo_multi(ma, path)
        _test_features_multi(ma, path)


def _update_expected_files(ma, path):
    # used only when test cases are added or modified
    _dump_all_multi(ma, path)
    if ma.global_config.simulations_filter_in_memory:
        ma2 = ma.apply_filter()
        _dump_all_multi(ma2, path / "_filtered")


def _copy_config_files(src_file, dst_dir):
    # copy the config file
    analysis_config_path = shutil.copy(src_file, dst_dir)
    # copy any not hidden subdirectory
    src_dir = src_file.parent
    with os.scandir(src_dir) as it:
        for entry in it:
            if not entry.name.startswith(".") and entry.is_dir():
                shutil.copytree(src_dir / entry.name, dst_dir / entry.name)
    return Path(analysis_config_path)


@pytest.mark.force_update
def test_update_expected_files(analysis_config_path, expected_path, tmp_path):
    np.random.seed(0)
    analysis_config_path = _copy_config_files(analysis_config_path, tmp_path)

    with MultiAnalyzer.from_file(analysis_config_path) as multi_analyzer:
        _update_expected_files(multi_analyzer, expected_path)


def test_analyzer(analysis_config_path, expected_path, tmp_path):
    np.random.seed(0)
    analysis_config_path = _copy_config_files(analysis_config_path, tmp_path)

    # test without cache
    with MultiAnalyzer.from_file(analysis_config_path) as multi_analyzer:
        _test_repo_multi(multi_analyzer, expected_path)
        _test_features_multi(multi_analyzer, expected_path)
        _test_filter_in_memory(multi_analyzer, expected_path)

    # test with cache
    with MultiAnalyzer.from_file(analysis_config_path) as multi_analyzer:
        _test_repo_multi(multi_analyzer, expected_path)
        _test_features_multi(multi_analyzer, expected_path)
        _test_filter_in_memory(multi_analyzer, expected_path)



from pathlib import Path
from unittest.mock import patch

from click.testing import CliRunner

from blueetl.apps import convert as test_module


@patch(test_module.__name__ + ".main")
def test_validate_config_success(mock_main, tmp_path):
    input_file = tmp_path / "spikes.csv"
    output_dir = tmp_path / "output"
    runner = CliRunner()

    with runner.isolated_filesystem(temp_dir=tmp_path) as td:
        input_file.write_text("ids times\n1 0.0")
        result = runner.invoke(
            test_module.convert_spikes,
            [
                str(input_file),
                str(output_dir),
                "--node-population",
                "custom",
            ],
        )

    assert result.output.strip() == "Conversion successful."
    assert result.exit_code == 0
    assert mock_main.call_count == 1



import textwrap

from click.testing import CliRunner

from blueetl import __version__
from blueetl.apps import main as test_module


def test_version():
    runner = CliRunner()
    result = runner.invoke(test_module.cli, ["--version"])

    assert result.output.strip() == f"cli, version {__version__}"
    assert result.exit_code == 0


def test_help():
    expected_message = """
    Usage: cli [OPTIONS] COMMAND [ARGS]...
    
      The CLI entry point.
    
    Options:
      --version  Show the version and exit.
      --help     Show this message and exit.
    
    Commands:
      run              Run the analysis.
      migrate-config   Migrate a configuration file.
      validate-config  Validate a configuration file.
      convert-spikes   Convert spikes in CSV format.
    """

    runner = CliRunner()
    result = runner.invoke(test_module.cli, ["--help"])

    assert result.output.strip() == textwrap.dedent(expected_message).strip()
    assert result.exit_code == 0



"""Analysis Configuration."""

import logging
from collections.abc import Iterator
from copy import deepcopy
from itertools import chain
from pathlib import Path
from typing import Any, NamedTuple, Optional, Union

from blueetl.config.analysis_model import (
    FeaturesConfig,
    MultiAnalysisConfig,
    TrialStepsConfig,
    WindowConfig,
)
from blueetl.constants import CHECKSUM_SEP
from blueetl.utils import checksum_json, dict_product, load_json
from blueetl.validation import read_schema, validate_config

L = logging.getLogger(__name__)


def _override_params(global_config: dict[str, Any], extra_params: dict[str, Any]) -> dict[str, Any]:
    """Override global_config with params in extra_params."""
    global_config = deepcopy(global_config)
    cache_config = global_config.setdefault("cache", {})
    if (value := extra_params.get("clear_cache")) is not None:
        cache_config["clear"] = value
    if (value := extra_params.get("readonly_cache")) is not None:
        cache_config["readonly"] = value
    if (value := extra_params.get("skip_features_cache")) is not None:
        cache_config["skip_features"] = value
    return global_config


def _resolve_paths(global_config: MultiAnalysisConfig, base_path: Path) -> None:
    """Resolve any relative path at the top level."""
    base_path = base_path or Path()
    global_config.simulation_campaign = base_path / global_config.simulation_campaign
    global_config.cache.path = base_path / global_config.cache.path


def _resolve_neuron_classes(global_config: MultiAnalysisConfig, base_path: Path):
    """Resolve the relative paths in neuron_classes."""
    for config in global_config.analysis.values():
        for neuron_classes_config in config.extraction.neuron_classes.values():
            if path := neuron_classes_config.node_sets_file:
                path = base_path / path
                neuron_classes_config.node_sets_file = path
                neuron_classes_config.node_sets_checksum = checksum_json(load_json(path))


def _resolve_trial_steps(global_config: MultiAnalysisConfig, base_path: Path):
    """Set trial_steps_config.base_path to the configuration base path.

    In this way, the custom function can use it as the base path to save any figure.
    """
    for config in global_config.analysis.values():
        for trial_steps_config in config.extraction.trial_steps.values():
            if not trial_steps_config.base_path:
                trial_steps_config.base_path = base_path
            if path := trial_steps_config.node_sets_file:
                path = base_path / path
                trial_steps_config.node_sets_file = path
                trial_steps_config.node_sets_checksum = checksum_json(load_json(path))


def _resolve_windows(global_config: MultiAnalysisConfig) -> None:
    """Calculate the hash of any referenced windows in each single analysis configuration.

    This is needed to invalidate the cache of the windows referring to the original windows.
    """

    class Reference(NamedTuple):
        """Reference attributes."""

        report: str
        window: str

    def _parse_ref(ref: str) -> Reference:
        # example: spikes.extraction.windows.w3 -> Reference(report="spikes", window="w3")
        try:
            report, _extraction, _windows, window = ref.split(".")
            if _extraction != "extraction" or _windows != "windows":
                raise ValueError
        except ValueError as ex:
            raise ValueError(
                "Windows reference is not in the expected format: "
                "<report_name>.extraction.windows.<window_name>"
            ) from ex
        return Reference(report, window)

    def _calculate_checksum(ref: str) -> str:
        reference = _parse_ref(ref)
        try:
            extraction = global_config.analysis[reference.report].extraction
            window = extraction.windows[reference.window]
        except KeyError as ex:
            raise ValueError(f"Windows reference {ref!r} points to nonexistent object") from ex
        assert isinstance(window, WindowConfig), "Invalid referenced window config"
        trial_step = None
        if trial_steps_label := window.trial_steps_label:
            L.debug("Checksum including trial_steps_label %s", trial_steps_label)
            # retrieve the dynamic trial_step config to be included in the checksum
            trial_step = extraction.trial_steps[trial_steps_label]
            assert isinstance(trial_step, TrialStepsConfig), "Invalid referenced trial_steps config"
        return checksum_json(
            [
                window.checksum(),
                trial_step.checksum() if trial_step else "",
            ]
        )

    for analysis_name, analysis_config in global_config.analysis.items():
        windows = analysis_config.extraction.windows
        new_windows: dict[str, Union[str, WindowConfig]] = {}
        for window_name, cfg in windows.items():
            if isinstance(cfg, str):
                checksum = _calculate_checksum(cfg)
                new_windows[window_name] = f"{cfg}{CHECKSUM_SEP}{checksum}"
                L.debug(
                    "Config checksum for analysis %s and window %s: %s",
                    analysis_name,
                    window_name,
                    checksum,
                )
            else:
                new_windows[window_name] = cfg
        analysis_config.extraction.windows = new_windows


def _resolve_features(features_config_list: list[FeaturesConfig]) -> list[FeaturesConfig]:
    def expand_product(params: dict, params_product: dict) -> Iterator[dict]:
        for items in dict_product(params_product):
            new_params = deepcopy(params)
            for key, v, i in items:
                new_params[key] = v
                new_params["__suffix__"] = new_params.get("__suffix__", "") + f"_{i}"
            yield new_params

    def expand_zip(params: dict, params_zip: dict) -> Iterator[dict]:
        if len({len(values) for values in params_zip.values()}) != 1:
            raise ValueError("All the zip params must have the same length")
        for i, zip_values in enumerate(zip(*params_zip.values())):
            new_params = deepcopy(params)
            new_params["__suffix__"] = new_params.get("__suffix__", "") + f"__{i}"
            new_params.update(zip(params_zip.keys(), zip_values))
            yield new_params

    new = []
    for config in features_config_list:
        config = deepcopy(config)
        params_list = [config.params]
        config.params = {}
        if params_product := config.params_product:
            config.params_product = {}
            params_list = list(
                chain.from_iterable(expand_product(p, params_product) for p in params_list)
            )
        if params_zip := config.params_zip:
            config.params_zip = {}
            params_list = list(chain.from_iterable(expand_zip(p, params_zip) for p in params_list))
        for p in params_list:
            new_features_config = deepcopy(config)
            new.append(new_features_config)
            if p:
                # suffix becomes part of the configuration
                new_features_config.suffix += p.pop("__suffix__", "")
                new_features_config.params = p
    return new


def _resolve_analysis_configs(global_config: MultiAnalysisConfig) -> None:
    for config in global_config.analysis.values():
        config.simulations_filter = global_config.simulations_filter
        config.simulations_filter_in_memory = global_config.simulations_filter_in_memory
        config.features = _resolve_features(config.features)


def init_multi_analysis_configuration(
    global_config: dict, base_path: Path, extra_params: Optional[dict[str, Any]]
) -> MultiAnalysisConfig:
    """Return a config object from a config dict."""
    global_config = _override_params(global_config, extra_params or {})
    validate_config(global_config, schema=read_schema("analysis_config"))
    config = MultiAnalysisConfig.model_validate(global_config)
    _resolve_paths(config, base_path=base_path)
    _resolve_neuron_classes(config, base_path=base_path)
    _resolve_trial_steps(config, base_path=base_path)
    _resolve_windows(config)
    _resolve_analysis_configs(config)
    return config



"""Convert CLI."""

import logging
from pathlib import Path

import click

from blueetl.converters.convert_spikes import main
from blueetl.utils import setup_logging


@click.command()
@click.argument("input-file", type=click.Path(exists=True, path_type=Path))
@click.argument("output-dir", type=click.Path(exists=False, path_type=Path))
@click.option(
    "--node-population",
    help="Name of the node population to create.",
    default="synthetic",
    show_default=True,
)
@click.option("-v", "--verbose", count=True, help="-v for INFO, -vv for DEBUG")
def convert_spikes(input_file, output_dir, node_population, verbose):
    """Convert spikes in CSV format.

    Read INPUT_FILE containing the spikes in CSV format, and write synthetic files to OUTPUT_DIR.

    The input file should contain:

    \b
    - headers: node_ids timestamps (or: ids times)
    - values: space separated
    """
    loglevel = (logging.WARNING, logging.INFO, logging.DEBUG)[min(verbose, 2)]
    setup_logging(loglevel=loglevel, force=True)
    main(
        input_file=input_file,
        output_dir=output_dir,
        node_population=node_population,
    )
    click.secho("Conversion successful.", fg="green")



import itertools
import os
from collections.abc import Iterator
from functools import partial
from types import SimpleNamespace
from typing import NamedTuple
from unittest.mock import Mock, patch

import pandas as pd
import pytest
from blueetl_core.constants import BLUEETL_JOBLIB_JOBS
from numpy.testing import assert_array_equal
from pandas.testing import assert_frame_equal

from blueetl import parallel as test_module


def merge_groupby_classic(
    df_list: list[pd.DataFrame], groupby: list[str]
) -> Iterator[tuple[NamedTuple, pd.DataFrame]]:
    """Merge a list of DataFrames, group by the given keys, and yield keys and groups."""
    merged = df_list[0]
    for df in df_list[1:]:
        merged = merged.merge(df, how="left")
    yield from merged.groupby(groupby, dropna=False)


def test_merge_groupby():
    all_dataframes = {
        "simulations": pd.DataFrame(
            [
                {
                    "ca": 1.1,
                    "seed": 1,
                    "simulation_path": "/path/to/0",
                    "simulation_id": 0,
                    "circuit_id": 0,
                    "simulation": Mock(),
                    "circuit": Mock(),
                },
                {
                    "ca": 1.2,
                    "seed": 1,
                    "simulation_path": "/path/to/1",
                    "simulation_id": 1,
                    "circuit_id": 0,
                    "simulation": Mock(),
                    "circuit": Mock(),
                },
                {
                    "ca": 1.1,
                    "seed": 1,
                    "simulation_path": "/path/to/2",
                    "simulation_id": 2,
                    "circuit_id": 1,
                    "simulation": Mock(),
                    "circuit": Mock(),
                },
                {
                    "ca": 1.2,
                    "seed": 1,
                    "simulation_path": "/path/to/3",
                    "simulation_id": 3,
                    "circuit_id": 1,
                    "simulation": Mock(),
                    "circuit": Mock(),
                },
            ]
        ),
        "neurons": pd.DataFrame(
            [
                {"circuit_id": 0, "neuron_class": "L2_INH", "gid": 0, "neuron_class_index": 0},
                {"circuit_id": 0, "neuron_class": "L2_INH", "gid": 1, "neuron_class_index": 1},
                {"circuit_id": 0, "neuron_class": "L2_EXC", "gid": 2, "neuron_class_index": 0},
                {"circuit_id": 0, "neuron_class": "L2_EXC", "gid": 3, "neuron_class_index": 1},
            ]
        ),
        "by_neuron_class": pd.DataFrame(
            [
                {"simulation_id": 0, "circuit_id": 0, "neuron_class": "L2_INH", "f1": 111},
                {"simulation_id": 0, "circuit_id": 0, "neuron_class": "L2_EXC", "f1": 222},
                {"simulation_id": 1, "circuit_id": 0, "neuron_class": "L2_INH", "f1": 333},
                {"simulation_id": 1, "circuit_id": 0, "neuron_class": "L2_EXC", "f1": 444},
                {"simulation_id": 2, "circuit_id": 1, "neuron_class": "L2_INH", "f1": 555},
                {"simulation_id": 2, "circuit_id": 1, "neuron_class": "L2_EXC", "f1": 666},
                {"simulation_id": 3, "circuit_id": 1, "neuron_class": "L2_INH", "f1": 777},
                {"simulation_id": 3, "circuit_id": 1, "neuron_class": "L2_EXC", "f1": 888},
            ]
        ),
    }
    df_list = list(all_dataframes.values())
    groupby = ["simulation_id", "circuit_id", "neuron_class"]

    result = test_module.merge_groupby(df_list, groupby=groupby)
    expected = merge_groupby_classic(df_list, groupby=groupby)

    for result_item, expected_item in itertools.zip_longest(result, expected):
        assert isinstance(result_item, tuple)
        assert isinstance(expected_item, tuple)
        result_key, result_group = result_item
        expected_key, expected_group = expected_item
        assert_array_equal(result_key, expected_key)
        assert isinstance(result_group, pd.DataFrame)
        assert isinstance(expected_group, pd.DataFrame)
        # Reset the index because the index is ignored in merge_groupby
        expected_group = expected_group.reset_index(drop=True)
        # dtype may be different because int64 can be converted to float64 when some values are nan
        assert_frame_equal(result_group, expected_group, check_dtype=False)


@pytest.mark.parametrize("how", ["namespace", "namedtuple", "dict", "series", "dataframe"])
@patch.dict(os.environ, {BLUEETL_JOBLIB_JOBS: "1"})
def test_call_by_simulation(how):
    simulations = pd.DataFrame(
        [
            {
                "ca": 1.1,
                "seed": 1,
                "simulation_path": "/path/to/0",
                "simulation_id": 0,
                "circuit_id": 0,
                "simulation": Mock(),
                "circuit": Mock(),
            },
            {
                "ca": 1.2,
                "seed": 1,
                "simulation_path": "/path/to/1",
                "simulation_id": 1,
                "circuit_id": 0,
                "simulation": Mock(),
                "circuit": Mock(),
            },
            {
                "ca": 1.1,
                "seed": 1,
                "simulation_path": "/path/to/2",
                "simulation_id": 2,
                "circuit_id": 1,
                "simulation": Mock(),
                "circuit": Mock(),
            },
            {
                "ca": 1.2,
                "seed": 1,
                "simulation_path": "/path/to/3",
                "simulation_id": 3,
                "circuit_id": 1,
                "simulation": Mock(),
                "circuit": Mock(),
            },
        ]
    )
    all_dataframes = {
        "neurons": pd.DataFrame(
            [
                {"circuit_id": 0, "neuron_class": "L2_INH", "gid": 0, "neuron_class_index": 0},
                {"circuit_id": 0, "neuron_class": "L2_INH", "gid": 1, "neuron_class_index": 1},
                {"circuit_id": 0, "neuron_class": "L2_EXC", "gid": 2, "neuron_class_index": 0},
                {"circuit_id": 0, "neuron_class": "L2_EXC", "gid": 3, "neuron_class_index": 1},
            ]
        ),
        "by_neuron_class": pd.DataFrame(
            [
                {"simulation_id": 0, "circuit_id": 0, "neuron_class": "L2_INH", "f1": 111},
                {"simulation_id": 0, "circuit_id": 0, "neuron_class": "L2_EXC", "f1": 222},
                {"simulation_id": 1, "circuit_id": 0, "neuron_class": "L2_INH", "f1": 333},
                {"simulation_id": 1, "circuit_id": 0, "neuron_class": "L2_EXC", "f1": 444},
                {"simulation_id": 2, "circuit_id": 1, "neuron_class": "L2_INH", "f1": 555},
                {"simulation_id": 2, "circuit_id": 1, "neuron_class": "L2_EXC", "f1": 666},
                {"simulation_id": 3, "circuit_id": 1, "neuron_class": "L2_INH", "f1": 777},
                {"simulation_id": 3, "circuit_id": 1, "neuron_class": "L2_EXC", "f1": 888},
            ]
        ).set_index(["simulation_id", "circuit_id", "neuron_class"]),
    }

    simulation_row_type = {
        "namespace": SimpleNamespace,
        "namedtuple": tuple,
        "dict": dict,
        "series": pd.Series,
        "dataframe": pd.DataFrame,
    }[how]
    simulation_row_get = {
        "namespace": lambda row, attr: getattr(row, attr),
        "namedtuple": lambda row, attr: getattr(row, attr),
        "dict": lambda row, attr: row[attr],
        "series": lambda row, attr: row.at[attr],
        "dataframe": lambda row, attr: row.at[0, attr],
    }[how]

    def func(simulation_row, filtered_dataframes, param1, param2):
        assert isinstance(simulation_row, simulation_row_type)
        assert isinstance(filtered_dataframes, dict)
        _df = filtered_dataframes["by_neuron_class"]
        assert isinstance(_df, pd.DataFrame)
        simulation_id = simulation_row_get(simulation_row, "simulation_id")
        assert_array_equal(_df.index.unique("simulation_id"), [simulation_id])
        return [simulation_id, sum(_df["f1"]), param1, param2]

    result = test_module.call_by_simulation(
        simulations=simulations,
        dataframes_to_filter=all_dataframes,
        func=partial(func, param1=1, param2=2),
        how=how,
    )

    assert result == [[0, 333, 1, 2], [1, 777, 1, 2], [2, 1221, 1, 2], [3, 1665, 1, 2]]



"""Soma Potentials extractor."""

import logging
from typing import Optional

import pandas as pd
from blueetl_core.utils import smart_concat

from blueetl.adapters.simulation import SimulationAdapter as Simulation
from blueetl.constants import CIRCUIT_ID, GID, NEURON_CLASS, SIMULATION_ID, TIME, VALUE, WINDOW
from blueetl.extract.report import ReportExtractor

L = logging.getLogger(__name__)


class SomaReport(ReportExtractor):
    """SomaReport extractor class."""

    COLUMNS = [SIMULATION_ID, CIRCUIT_ID, NEURON_CLASS, WINDOW, TIME, GID, VALUE]

    @classmethod
    def _load_values(
        cls,
        simulation: Simulation,
        population: Optional[str],
        gids,
        windows_df: pd.DataFrame,
        name: str,
    ) -> pd.DataFrame:
        """Return a DataFrame for the given simulation, population, gids, and windows."""
        df_list = []
        report = simulation.reports[name][population]
        for rec in windows_df.itertuples():
            win = cls.calculate_window_slice(rec)
            df = report.get(group=gids, t_start=win.t_start, t_stop=win.t_stop, t_step=win.t_step)
            df.index.rename(TIME, inplace=True)
            df.columns.rename(GID, inplace=True)
            df = df.unstack().reset_index()
            df.rename(columns={0: VALUE}, inplace=True)
            df[WINDOW] = win.name
            df_list.append(df)
        return smart_concat(df_list)



import contextlib
from pathlib import Path

TEST_DATA_PATH = Path(__file__).parent / "data"


@contextlib.contextmanager
def assertion_error_message(msg):
    try:
        yield
    except AssertionError as ex:
        raise AssertionError(f"{msg}\n{ex}") from ex



from click.testing import CliRunner

from blueetl.apps import migrate as test_module
from blueetl.constants import CONFIG_VERSION
from blueetl.utils import dump_yaml, load_yaml


def test_supported_version():
    assert (
        test_module.MIGRATION_CONFIG_VERSION == CONFIG_VERSION
    ), "The migration script needs to be updated to support the current configuration version"


def test_migrate_config(tmp_path):
    input_config_file = "input_config.yaml"
    output_config_file = "output_config.yaml"
    input_data = {
        "simulation_campaign": "/path/to/config.json",
        "simulations_filter_in_memory": {"simulation_id": 2},
        "output": "output_dir",
        "extraction": {
            "neuron_classes": {
                "L1_EXC": {"layer": ["1"], "synapse_class": ["EXC"]},
                "L1_EXC_gids": {"layer": ["1"], "synapse_class": ["EXC"], "gid": [1, 2]},
            },
            "limit": None,
            "target": None,
            "windows": {"w1": {"bounds": [20, 90], "window_type": "spontaneous"}},
        },
        "analysis": {
            "features": [
                {
                    "type": "multi",
                    "groupby": ["simulation_id", "circuit_id", "neuron_class", "window"],
                    "function": "module.user.function",
                    "params": {"export_all_neurons": True},
                }
            ]
        },
    }
    expected_data = {
        "version": CONFIG_VERSION,
        "simulation_campaign": "/path/to/config.json",
        "simulations_filter_in_memory": {"simulation_id": 2},
        "cache": {"path": "output_dir"},
        "analysis": {
            "spikes": {
                "extraction": {
                    "report": {"type": "spikes"},
                    "neuron_classes": {
                        "L1_EXC": {"query": {"layer": ["1"], "synapse_class": ["EXC"]}},
                        "L1_EXC_gids": {
                            "query": {"layer": ["1"], "synapse_class": ["EXC"]},
                            "node_id": [1, 2],
                        },
                    },
                    "limit": None,
                    "node_set": None,
                    "windows": {"w1": {"bounds": [20, 90], "window_type": "spontaneous"}},
                },
                "features": [
                    {
                        "type": "multi",
                        "groupby": ["simulation_id", "circuit_id", "neuron_class", "window"],
                        "function": "module.user.function",
                        "params": {"export_all_neurons": True},
                    }
                ],
            }
        },
    }
    expected_message = "The converted configuration has been saved to output_config.yaml."
    runner = CliRunner()

    with runner.isolated_filesystem(temp_dir=tmp_path):
        dump_yaml(input_config_file, input_data)
        result = runner.invoke(test_module.migrate_config, [input_config_file, output_config_file])
        assert result.exit_code == 0
        assert result.output.strip() == expected_message
        output_data = load_yaml(output_config_file)

    assert output_data == expected_data



"""Analysis functions."""

import gc
import logging
from pathlib import Path
from typing import Any, NamedTuple, Optional

import numpy as np
import pandas as pd

from blueetl.cache import CacheManager
from blueetl.campaign.config import SimulationCampaign
from blueetl.config.analysis import init_multi_analysis_configuration
from blueetl.config.analysis_model import CacheConfig, MultiAnalysisConfig, SingleAnalysisConfig
from blueetl.features import FeaturesCollection
from blueetl.repository import Repository
from blueetl.resolver import AttrResolver, Resolver
from blueetl.types import StrOrPath
from blueetl.utils import load_yaml, setup_logging

L = logging.getLogger(__name__)


class Analyzer:
    """Analyzer class."""

    def __init__(
        self,
        analysis_config: SingleAnalysisConfig,
        repo: Repository,
        features: FeaturesCollection,
    ) -> None:
        """Initialize the object.

        Args:
            analysis_config: analysis configuration.
            repo: Repository instance.
            features: FeaturesCollection instance.
        """
        self._analysis_config = analysis_config
        self._repo = repo
        self._features = features

    @classmethod
    def from_config(
        cls,
        analysis_config: SingleAnalysisConfig,
        simulations_config: SimulationCampaign,
        cache_config: CacheConfig,
        resolver: Resolver,
    ) -> "Analyzer":
        """Initialize the Analyzer from the given configuration.

        Args:
            analysis_config: analysis configuration.
            simulations_config: simulation campaign configuration.
            cache_config: cache configuration.
            resolver: resolver instance.
        """
        cache_manager = CacheManager(
            cache_config=cache_config,
            analysis_config=analysis_config,
            simulations_config=simulations_config,
        )
        repo = Repository(
            simulations_config=simulations_config,
            extraction_config=analysis_config.extraction,
            cache_manager=cache_manager,
            simulations_filter=analysis_config.simulations_filter,
            resolver=resolver,
        )
        features = FeaturesCollection(
            features_configs=analysis_config.features,
            repo=repo,
            cache_manager=cache_manager,
        )
        return cls(
            analysis_config=analysis_config,
            repo=repo,
            features=features,
        )

    def __enter__(self):
        """Initialize the object when used as a context manager."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Finalize the object when used as a context manager."""
        self.close()

    def close(self) -> None:
        """Invalidate and unlock the cache.

        After calling this method, the DataFrames already extracted can still be accessed,
        but it's not possible to extract new data or calculate new features.
        """
        self.repo.cache_manager.close()
        self.features.cache_manager.close()

    @property
    def analysis_config(self) -> SingleAnalysisConfig:
        """Return the wrapped analysis configuration."""
        return self._analysis_config

    @property
    def repo(self) -> Repository:
        """Return the wrapped repository."""
        return self._repo

    @property
    def extraction(self) -> Repository:
        """Return the wrapped repository as an alias."""
        return self._repo

    @property
    def features(self) -> FeaturesCollection:
        """Return the wrapped features."""
        return self._features

    def extract_repo(self) -> None:
        """Extract all the repositories dataframes."""
        self.repo.extract()

    def calculate_features(self) -> None:
        """Calculate all the features defined in the configuration."""
        self.features.calculate()
        gc.collect()

    def apply_filter(self, simulations_filter: Optional[dict[str, Any]] = None) -> "Analyzer":
        """Return a new object where the in memory filter is applied to repo and features.

        Before applying the filter, all the repo dataframes are extracted,
        and all the features dataframes are calculated, if not already done.

        Args:
            simulations_filter: optional simulations filter; if not specified, use
                `simulations_filter_in_memory` from the configuration; if neither is specified,
                return the original object.
        """
        self.extract_repo()
        self.calculate_features()
        if not simulations_filter:
            simulations_filter = self.analysis_config.simulations_filter_in_memory
        if not simulations_filter:
            return self
        repo = self.repo.apply_filter(simulations_filter)
        features = self.features.apply_filter(repo)
        return Analyzer(analysis_config=self.analysis_config, repo=repo, features=features)

    def show(self):
        """Print all the DataFrames."""
        self.repo.show()
        self.features.show()

    def try_one(self, groupby: list[str]) -> tuple[NamedTuple, pd.DataFrame]:
        """Return the first key and df when grouping spikes by the given list of columns.

        The returned parameters are the same passed to the feature function defined by the user.

        It should be used only for internal use and debug.

        Args:
            groupby: list of columns to group by.

        Returns:
            The first key and df.
        """
        key, df = next(self.repo.report.df.etl.groupby_iter(groupby))
        return key, df


class MultiAnalyzer:
    """MultiAnalyzer class."""

    def __init__(
        self,
        global_config: MultiAnalysisConfig,
        analyzers: Optional[dict[str, Analyzer]] = None,
    ) -> None:
        """Initialize the object.

        Args:
            global_config: analysis configuration.
            analyzers: dict of analyzers, or None to load them from the configuration.
        """
        assert isinstance(global_config, MultiAnalysisConfig)
        self._global_config = global_config
        self._analyzers = self._init_analyzers() if analyzers is None else analyzers

    @classmethod
    def from_config(
        cls,
        global_config: dict,
        base_path: StrOrPath,
        extra_params: Optional[dict[str, Any]] = None,
    ) -> "MultiAnalyzer":
        """Initialize the MultiAnalyzer from the given configuration.

        Args:
            global_config: analysis configuration.
            base_path: base path used to resolve relative paths in the configuration.
            extra_params: dict of overriding parameters.
        """
        global_config = init_multi_analysis_configuration(
            global_config, base_path=Path(base_path), extra_params=extra_params
        )
        return cls(global_config=global_config)

    def _init_analyzers(self) -> dict[str, Analyzer]:
        """Load and return a dict of analyzers."""
        resolver = AttrResolver(root=self)
        simulations_config = SimulationCampaign.load(self.global_config.simulation_campaign)
        return {
            name: Analyzer.from_config(
                analysis_config=analysis_config,
                simulations_config=simulations_config,
                cache_config=self.global_config.cache.model_copy(
                    update={"path": self.global_config.cache.path / name}
                ),
                resolver=resolver,
            )
            for name, analysis_config in self.global_config.analysis.items()
        }

    @classmethod
    def from_file(
        cls, path: StrOrPath, extra_params: Optional[dict[str, Any]] = None
    ) -> "MultiAnalyzer":
        """Return a new instance loaded using the given configuration file."""
        return cls.from_config(
            global_config=load_yaml(path),
            base_path=Path(path).parent,
            extra_params=extra_params,
        )

    @property
    def global_config(self) -> MultiAnalysisConfig:
        """Return the global config instance."""
        return self._global_config

    @property
    def analyzers(self) -> dict[str, Analyzer]:
        """Return the dict of analyzers."""
        return self._analyzers

    @property
    def names(self) -> list[str]:
        """Return the names of all the analyzers."""
        return list(self.analyzers)

    def __getstate__(self) -> dict:
        """Get the object state when the object is pickled."""
        return {"_global_config": self.global_config, "_analyzers": None}

    def __setstate__(self, state: dict) -> None:
        """Set the object state when the object is unpickled."""
        self.__dict__.update(state)

    def __getattr__(self, name: str) -> Analyzer:
        """Return an analyzer instance by name.

        Args:
            name: name of the analyzer.
        """
        try:
            return self.analyzers[name]
        except KeyError as ex:
            raise AttributeError(
                f"{self.__class__.__name__!r} object has no attribute {name!r}"
            ) from ex

    def __dir__(self):
        """Allow autocompletion of dynamic attributes."""
        return list(super().__dir__()) + list(self.analyzers)

    def __enter__(self):
        """Initialize the object when used as a context manager."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Finalize the object when used as a context manager."""
        self.close()

    def close(self) -> None:
        """Invalidate and unlock the cache.

        After calling this method, the DataFrames already extracted can still be accessed,
        but it's not possible to extract new data or calculate new features.
        """
        for a in self.analyzers.values():
            a.close()

    def extract_repo(self) -> None:
        """Extract all the repositories dataframes for all the analysis."""
        for a in self.analyzers.values():
            a.extract_repo()

    def calculate_features(self) -> None:
        """Calculate all the features defined in the configuration for all the analysis."""
        for a in self.analyzers.values():
            a.calculate_features()

    def apply_filter(self, simulations_filter: Optional[dict[str, Any]] = None) -> "MultiAnalyzer":
        """Return a new object where the in memory filter is applied to repo and features.

        Before applying the filter, all the repo dataframes are extracted,
        and all the features dataframes are calculated, if not already done.

        Args:
            simulations_filter: optional simulations filter; if not specified, use
                `simulations_filter_in_memory` from the configuration; if neither is specified,
                return the original object.
        """
        self.extract_repo()
        self.calculate_features()
        if not simulations_filter:
            simulations_filter = self.global_config.simulations_filter_in_memory
        if not simulations_filter:
            return self
        analyzers = {name: a.apply_filter(simulations_filter) for name, a in self.analyzers.items()}
        return MultiAnalyzer(
            global_config=self.global_config.model_copy(deep=True),
            analyzers=analyzers,
        )

    def show(self):
        """Print all the DataFrames."""
        for name in self.names:
            print("#" * 80)
            print("Analysis:", name)
            getattr(self, name).show()


def run_from_file(
    analysis_config_file: StrOrPath,
    seed: Optional[int] = 0,
    extract: bool = True,
    calculate: bool = True,
    show: bool = False,
    clear_cache: Optional[bool] = None,
    readonly_cache: Optional[bool] = None,
    skip_features_cache: Optional[bool] = None,
    loglevel: Optional[int] = None,
) -> MultiAnalyzer:
    """Initialize and return the MultiAnalyzer.

    Args:
        analysis_config_file: path to the analysis configuration file.
        seed: if not None, random seed used to select random gids.
        extract: if True, run the extraction of the repository.
        calculate: if True, run the calculation of the features.
        show: if True, show a short representation of all the Pandas DataFrames, mainly for debug.
        clear_cache: if None, use the value from the configuration file. Otherwise:
            if True, remove any existing cache;
            if False, reuse the existing cache if possible.
        readonly_cache: if None, use the value from the configuration file. Otherwise:
            if True, use the existing cache if possible, or raise an error;
            if False, use the existing cache if possible, or update it.
        skip_features_cache:  if None, use the value from the configuration file. Otherwise:
            if True, do not write the features to the cache;
            if False, write the features to the cache after calculating them.
        loglevel: if specified, used to set up logging.

    Returns:
        a new MultiAnalyzer instance.
    """
    # pylint: disable=too-many-arguments
    if loglevel is not None:
        setup_logging(loglevel=loglevel, force=True)
    if seed is not None:
        np.random.seed(seed)
    L.info("MultiAnalyzer configuration: %s", analysis_config_file)
    ma = MultiAnalyzer.from_file(
        analysis_config_file,
        extra_params={
            "clear_cache": clear_cache,
            "readonly_cache": readonly_cache,
            "skip_features_cache": skip_features_cache,
        },
    )
    if extract:
        ma.extract_repo()
    if calculate:
        ma.calculate_features()
    if show:
        ma.show()
    return ma



"""Validate CLI."""

import sys

import click

from blueetl import validation
from blueetl.utils import load_yaml


@click.command()
@click.argument("analysis_config_file", type=click.Path(exists=True))
def validate_config(analysis_config_file):
    """Validate a configuration file."""
    analysis_config = load_yaml(analysis_config_file)
    try:
        validation.validate_config(
            analysis_config, schema=validation.read_schema("analysis_config")
        )
    except validation.ValidationError:
        click.secho("Validation failed.", fg="red")
        sys.exit(1)
    else:
        click.secho("Validation successful.", fg="green")



"""neuron_class features from BlueCV."""

import logging
from functools import partial

import numpy as np
import pandas as pd
from elephant import spectral, statistics
from elephant.spike_train_correlation import correlation_coefficient
from quantities import ms

from blueetl.constants import GID, TIME
from blueetl.external.bluecv.utils import to_binned_spiketrain, to_spiketrains

L = logging.getLogger(__name__)


def calculate_features_by_neuron_class(repo, key, df, params):
    """Calculate features grouped by neuron_class."""
    t_start, t_stop = repo.windows.get_bounds(key.window)
    # create an array containing multiple arrays of spikes, one for each gid
    spiketrains = df.groupby([GID])[TIME].apply(np.array).to_numpy()
    ST = to_spiketrains(spiketrains, t_start, t_stop)
    BST = to_binned_spiketrain(ST)
    functions = {
        "PSD": partial(get_PSD, spiketrains),
        "AC": partial(get_AC, BST),
        "CPDF": partial(get_CPDF, ST),
        "PSTH": partial(get_PSTH, spiketrains, t_start=t_start, t_stop=t_stop),
    }
    result = {}
    for feature_name, feature_config in params.items():
        feature_params = feature_config.get("params", {})
        result[feature_name] = [functions[feature_name](**feature_params)]
    return {"by_neuron_class": pd.DataFrame(result)}


def get_PSD(spiketrains, n_segments=2):
    """Get power spectrum density of neuronal population."""
    _, PSD = spectral.welch_psd(
        np.concatenate(spiketrains),
        n_segments=n_segments,
    )
    return PSD


def get_AC(BST):
    """Get mean firing rate of neuronal population."""
    return np.triu(correlation_coefficient(BST), k=1)


def get_CPDF(ST, bin_size=20):
    """Get complexity."""
    CPDF = statistics.Complexity(spiketrains=ST, bin_size=bin_size * ms).pdf()
    return np.asarray(CPDF).reshape(-1)


def get_PSTH(spiketrains, t_start, t_stop, bin_size=20):
    """Get perstimulus time histogram of the selected population.

    Args:
        spiketrains: array containing multiple arrays of spikes
        t_start (float): start time
        t_stop (float): stop time
        bin_size (float): bin size in ms

    Returns:
        a list with two items:
            counts (np.ndarray): number of spike_times per bin in psth
            bins (np.ndarray): start of the bins
    """
    spiketrains = np.concatenate(spiketrains)
    inputbins = np.arange(t_start, t_stop, bin_size)
    counts, bins = np.histogram(
        spiketrains,
        bins=inputbins.size,
        range=(t_start, t_stop),
    )
    return [counts, bins[:-1]]



"""Migrate CLI."""

import sys
from copy import deepcopy

import click

from blueetl import validation
from blueetl.constants import CONFIG_VERSION
from blueetl.utils import dump_yaml, load_yaml

# it should always match CONFIG_VERSION if the script supports the current version
MIGRATION_CONFIG_VERSION = 4


def _safe_set(d, key, value):
    """Set the value in the given dict, or raise if it already exists."""
    if key in d:
        raise RuntimeError(f"The key {key} is already present in {d}")
    d[key] = value


def _rename_key(d, old, new):
    """Rename a key in the given dict, only if the old key already exists."""
    if old in d:
        _safe_set(d, new, d.pop(old))


def _migrate_v1_to_v2(input_config):
    """Migrate the configuration from v1 (BlueETL 0.1.x) to v2 (BlueETL 0.2.x)."""

    def _process_extraction(extraction):
        for neuron_class in extraction["neuron_classes"].values():
            _rename_key(neuron_class, "gid", "$gids")
        return extraction

    input_config = deepcopy(input_config)
    return {
        "version": 2,
        **{
            k: input_config.pop(k)
            for k in [
                "simulation_campaign",
                "output",
                "simulations_filter",
                "simulations_filter_in_memory",
            ]
            if k in input_config
        },
        "analysis": {
            "spikes": {
                "extraction": {
                    "report": {"type": "spikes"},
                    **_process_extraction(input_config.pop("extraction")),
                },
                "features": input_config.pop("analysis", {}).pop("features", []),
                **({"custom": input_config} if input_config else {}),
            }
        },
    }


def _migrate_v2_to_v3(input_config):
    """Migrate the configuration from v2 (BlueETL 0.2.x) to v3 (BlueETL 0.3.x)."""
    output_config = deepcopy(input_config)
    output_config["version"] = 3
    for analysis in output_config["analysis"].values():
        extraction = analysis["extraction"]
        _rename_key(extraction, "target", "node_set")
        for neuron_class in extraction["neuron_classes"].values():
            query = {}
            for k in list(neuron_class):
                if not k.startswith("$"):
                    query[k] = neuron_class.pop(k)
            if query:
                _safe_set(neuron_class, "query", query)
            _rename_key(neuron_class, "$target", "node_set")
            _rename_key(neuron_class, "$limit", "limit")
            _rename_key(neuron_class, "$gids", "node_id")
            _rename_key(neuron_class, "$query", "query")
    return output_config


def _migrate_v3_to_v4(input_config):
    """Migrate the configuration from v3 (BlueETL 0.8.x) to v4 (BlueETL 0.9.x)."""
    output_config = deepcopy(input_config)
    output_config["version"] = 4
    cache_config = output_config.setdefault("cache", {})
    if (value := output_config.pop("output", None)) is not None:
        _safe_set(cache_config, "path", value)
    if (value := output_config.pop("clear_cache", None)) is not None:
        _safe_set(cache_config, "clear", value)
    return output_config


def _sort_root_keys(input_config):
    root_keys = [
        "version",
        "simulation_campaign",
        "simulations_filter",
        "simulations_filter_in_memory",
        "cache",
        "analysis",
        "custom",
    ]
    output_config = {k: input_config.pop(k) for k in root_keys if k in input_config}
    return output_config | input_config


@click.command()
@click.option("--sort/--no-sort", help="Sort the root keys.", default=True, show_default=True)
@click.argument("input_config_file", type=click.Path(exists=True))
@click.argument("output_config_file", type=click.Path(exists=False))
def migrate_config(input_config_file, output_config_file, sort):
    """Migrate a configuration file."""
    config = load_yaml(input_config_file)
    version = config.get("version", 1)
    if not isinstance(version, int) or version < 1 or version > CONFIG_VERSION:
        click.secho(f"The config version {version} isn't supported.", fg="red")
        sys.exit(1)
    if version <= 1:
        config = _migrate_v1_to_v2(config)
    if version <= 2:
        config = _migrate_v2_to_v3(config)
    if version <= 3:
        config = _migrate_v3_to_v4(config)
    if version == CONFIG_VERSION:
        click.secho(f"The config version {version} doesn't need to be migrated.", fg="yellow")
    if sort:
        config = _sort_root_keys(config)
    validation.validate_config(config, schema=validation.read_schema("analysis_config"))
    dump_yaml(output_config_file, config, default_style="", default_flow_style=None)
    click.secho(f"The converted configuration has been saved to {output_config_file}.", fg="green")



"""Run CLI."""

import logging
import sys

import click

from blueetl import __version__
from blueetl.analysis import run_from_file


@click.command()
@click.argument("analysis_config_file", type=click.Path(exists=True))
@click.option("--seed", type=int, default=0, help="Pseudo-random generator seed", show_default=True)
@click.option("--extract/--no-extract", help="Extract (or load from the cache) the repository.")
@click.option("--calculate/--no-calculate", help="Calculate (or load from the cache) the features.")
@click.option("--show/--no-show", help="Show repository and features dataframes.")
@click.option(
    "--clear-cache/--no-clear-cache",
    help="If True, force clearing the cache.",
    default=None,
)
@click.option(
    "--readonly-cache/--no-readonly-cache",
    help="If True, use the existing cache if possible, or raise an error if not.",
    default=None,
)
@click.option(
    "--skip-features-cache/--no-skip-features-cache",
    help="If True, do not write the features to the cache.",
    default=None,
)
@click.option("-i", "--interactive/--no-interactive", help="Start an interactive IPython shell.")
@click.option("-v", "--verbose", count=True, help="-v for INFO, -vv for DEBUG")
def run(
    analysis_config_file,
    seed,
    extract,
    calculate,
    show,
    clear_cache,
    readonly_cache,
    skip_features_cache,
    interactive,
    verbose,
):
    """Run the analysis."""
    # pylint: disable=unused-variable,unused-import,import-outside-toplevel,too-many-arguments,too-many-locals
    loglevel = (logging.WARNING, logging.INFO, logging.DEBUG)[min(verbose, 2)]
    # assign the result to a local variable to make it available in the interactive shell
    ma = run_from_file(  # noqa
        analysis_config_file=analysis_config_file,
        seed=seed,
        extract=extract,
        calculate=calculate,
        show=show,
        clear_cache=clear_cache,
        readonly_cache=readonly_cache,
        skip_features_cache=skip_features_cache,
        loglevel=loglevel,
    )
    if interactive:
        # make np and pd immediately available in the interactive shell
        import numpy as np  # noqa
        import pandas as pd  # noqa

        try:
            from IPython import embed
        except ImportError:
            click.secho("You need to install IPython to start an interactive session.", fg="red")
            sys.exit(1)
        header = (
            f"BlueETL Interactive Shell {__version__}\n\n"
            f"The loaded MultiAnalyzer instance is available for test in the variable 'ma'."
        )
        embed(header=header, colors="neutral")



"""HDF data store."""

import logging
from typing import Optional

import pandas as pd

from blueetl.store.base import BaseStore
from blueetl.utils import timed

L = logging.getLogger(__name__)


def _category_to_object(df: pd.DataFrame) -> pd.DataFrame:
    # Prevent error: Cannot store a category dtype in a HDF5 dataset that uses format="fixed"
    return df.astype({name: "object" for name in df.columns[df.dtypes == "category"]})


class HDFStore(BaseStore):
    """HDF data store."""

    @property
    def extension(self) -> str:
        """Return the file extension to be used with this specific data store."""
        return "h5"

    def dump(self, df: pd.DataFrame, name: str) -> None:
        """Save a dataframe to file, using the given name and the class extension."""
        path = self.path(name)
        with timed(L.debug, f"Writing {name} to {path}"):
            df = _category_to_object(df)
            df.to_hdf(
                str(path),
                key=name,
                mode="w",
                # complib="blosc",
                # complevel=9,
                # format="fixed",
                # format="table",
            )

    def load(self, name: str) -> Optional[pd.DataFrame]:
        """Load a dataframe from file, using the given name and the class extension."""
        path = self.path(name)
        if not path.exists():
            return None
        with timed(L.debug, f"Reading {name} from {path}"):
            return pd.read_hdf(path, key=name)



import pytest

from blueetl.campaign.config import SimulationCampaign
from blueetl.config.analysis import init_multi_analysis_configuration
from blueetl.config.analysis_model import FeaturesConfig
from blueetl.features import FeaturesCollection
from blueetl.repository import Repository
from blueetl.utils import load_yaml
from tests.unit.utils import TEST_DATA_PATH, PicklableMock


@pytest.fixture
def lazy_fixture(request):
    """Return a function that returns the actual fixture from its name."""
    return lambda name: request.getfixturevalue(name)


@pytest.fixture
def global_config(tmp_path):
    config_path = TEST_DATA_PATH / "analysis" / "analysis_config_01_relative.yaml"
    config_dict = load_yaml(config_path)
    config_dict["cache"]["path"] = str(tmp_path / "analysis_output")
    return init_multi_analysis_configuration(config_dict, config_path.parent, extra_params={})


@pytest.fixture
def repo(global_config):
    simulations_config = SimulationCampaign.load(global_config.simulation_campaign)
    extraction_config = global_config.analysis["spikes"].extraction
    cache_manager = PicklableMock(
        is_repo_cached=PicklableMock(return_value=False),
        load_repo=PicklableMock(return_value=None),
        load_features=PicklableMock(return_value=None),
        get_cached_features_checksums=PicklableMock(return_value={}),
    )
    simulations_filter = global_config.simulations_filter
    resolver = PicklableMock()

    return Repository(
        simulations_config=simulations_config,
        extraction_config=extraction_config,
        cache_manager=cache_manager,
        simulations_filter=simulations_filter,
        resolver=resolver,
    )


@pytest.fixture
def features(repo):
    features_configs = [
        FeaturesConfig(
            type="multi",
            groupby=["simulation_id", "circuit_id", "neuron_class", "window"],
            function="blueetl.external.bnac.calculate_features.calculate_features_multi",
            params={"export_all_neurons": True},
            neuron_classes=["L2_X", "L6_Y"],
            windows=["w0", "w1"],
        )
    ]
    return FeaturesCollection(
        features_configs=features_configs,
        repo=repo,
        cache_manager=repo.cache_manager,
    )


@pytest.fixture
def features_with_suffixes(repo):
    features_configs = [
        FeaturesConfig(
            type="multi",
            groupby=["simulation_id", "circuit_id", "neuron_class", "window"],
            function="blueetl.external.bnac.calculate_features.calculate_features_multi",
            params={"export_all_neurons": True},
            neuron_classes=["L2_X", "L6_Y"],
            windows=["w0", "w1"],
            suffix="_0",
        ),
        FeaturesConfig(
            type="multi",
            groupby=["simulation_id", "circuit_id", "neuron_class", "window"],
            function="blueetl.external.bnac.calculate_features.calculate_features_multi",
            params={"export_all_neurons": False},
            neuron_classes=["L2_X", "L6_Y"],
            windows=["w0", "w1"],
            suffix="_1",
        ),
    ]
    return FeaturesCollection(
        features_configs=features_configs,
        repo=repo,
        cache_manager=repo.cache_manager,
    )
